#include "RegExpr.h"
#include "util/FastHash.h"

// @@@@ Some operations, especially those related to keeping REs in a 
// @@@@ near-canonical form, may be expensive yet be repeatedly applied. To
// @@@@ make them more efficient, use first() and/or minlen() to quickly
// @@@@ prune the search space.

// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Controlling nondeterminism and space explosion: first, we need to use
// the right size in CondFac that includes regexp size, instead of counting
// all regexp as one. Second, we need to devise the right way to introduce 
// nondeterminism. (But first try if the current one will be any good!)
// One key is to avoid expanding out *s if that is going to introduce too
// much overlap. (This may actually be a problem posed by the interation of
// alternation from different regexps, and may not be a real issue of star.)
// The second is that ND links, instead of going back to an ancestor state
// with the input pointer reset to what it was before, we can try to go forward 
// at compile time from that link, much like how failure links are constructed
// in Aho-Corasick.

RegExpr::
RegExpr(const CharClass& cc): op_(CLASS), class_() {
   const string& chars = cc.chars();
   if ((chars.size() == 0 && !cc.negated()) || 
       (chars.size()>=REGEXP_CHMAX-1 && cc.negated()))
      op_ = EMPTY;
   else {
      for (unsigned i=0; i < chars.size(); i++)
         class_.set((unsigned char)chars[i]);
      if (cc.negated()) class_.flip();
   }
   assert(isCanonical());
}

RegExpr::
RegExpr(const string& s): arg_() {
   unsigned sl = s.size();
   if (sl == 0) 
      new (this) RegExpr(END);
   else {
      op_ = SEQ;
      arg_.emplace_back(s[0]);
      arg_.emplace_back(s.substr(1));
   }
}

RegExpr::
RegExpr(RegExpOp op, const RegExpr* arg1, const RegExpr* arg2): arg_() {
   // We try to perform some reductions during construction. The higher level
   // goal is to get to something like a canonical form so that equal regexpr
   // are more likely to be syntactically equal. Obviously, we cannot get full
   // canonical form, as the underlying problems are too expensive to solve. 
   //
   // The REAL goal is to get a form that is canonical enough to recognize and
   // avoid redundancies that occur when residue operation is applied. Critical
   // in this regard is to keep ALT in flattened form without syntactically
   // identical duplicates. Once this is done, the residue-based Brzozowski's
   // algorithm won't generate any more syntactically distinct residues than
   // what you get with McNaughton-Yamada algorithm.
   //
   op_ = op;
   switch (op_) {
   case FALSE:
   case TRUE:
   case EMPTY:
   case BEG:
      assert(arg1 == nullptr && arg2 == nullptr);
      break;

   case END:
      assert(arg1 == nullptr && arg2 == nullptr);
      op_ = CLASS;
      class_ = bitset<REGEXP_CHMAX>();
      class_.set(0);
      break;

   case STAR:
      switch(arg1->op_) {
      case FALSE:
      case TRUE:
      case EMPTY:
      case BEG:
         op_ = arg1->op_;
         break;

      case CLASS:
         if (arg1->class_.count() >= REGEXP_CHMAX-1) // Sigma* == TRUE
            op_ = TRUE;
         else arg_.emplace_back(*arg1);
         break;

      case STAR: // x** = x*
         arg_.emplace_back(arg1->arg_[0]);
         break;

      case ALT: {
         arg_.emplace_back(*arg1);
         for (RegExpr& re: arg_)
            if (re.op_ == STAR) // (r1|...|rk*|...|rn)* = (r1|...|rk|...|rn)*
               re = re.arg_[0];
         break;
      }

      default:
         arg_.push_back(*arg1);
         break;
      }
      break;

   case PLUS: {
      RegExpr tre(STAR, arg1);
      RegExpr re(SEQ, arg1, &tre);
      *this = re;
      break;
   }

   case QMARK: {
      RegExpr tre(EMPTY);
      RegExpr re(ALT, arg1, &tre);
      *this = re;
      break;
   }

   case SEQ:
      if (arg1->op_ == FALSE || arg2->op_ == FALSE)
         op_ = FALSE;
      else if (arg1->op_ == TRUE && arg2->inclEpsilon())
         op_ = TRUE;
      else if (arg1->isEpsilon())
         *this = *arg2;
      else if (arg2->isEpsilon())
         *this = *arg1;
      else {
         arg_.emplace_back(*arg1);
         arg_.emplace_back(*arg2);
      }
      break;

   case ALT:
      *this = *arg1;
      if (arg2)
         merge(*arg2);
      break;

   default:
      assert(false);
   }
   assert(isCanonical());
};

void RegExpr::
merge(const RegExpr& re) {
   if (isInstance(re))
      *this = re;
   else if (re.isInstance(*this))
      return;
   else if (op_ == CLASS && re.op_ == CLASS)
      class_ |= re.class_;
   else if (op_ == ALT) {
      if (re.op_ == ALT) {
         for (const RegExpr& r: re.arg_) {
            for (const RegExpr& t: arg_) 
               if (r.isInstance(t))
                  goto skip;
            arg_.push_back(r);
         skip: continue;
         }
      }
      else if (!re.isInstance(*this))
         arg_.push_back(re);
   }
   else {
      RegExpr r = *this;
      if (op_ < STAR) 
         new (&arg_) vector<RegExpr>();
      op_ = ALT;
      arg_.clear();
      arg_.push_back(r);
      arg_.push_back(re);
   }
}


bool RegExpr::
isCanonical() const {
#ifndef CHECK_CANONICAL
   return true;
#endif
   switch (op_) {
   case FALSE: 
   case TRUE:
   case EMPTY:
      return true;

   case END:
      assert(false);

   case BEG:
   case PLUS:
   case QMARK:
      return false;

   case CLASS: 
      return (class_.count() > 0);

   case STAR:
      if (!arg_[0].isCanonical()) return false;
      switch (arg_[0].op_) {
      case TRUE: case FALSE: case EMPTY: case BEG: case END: case STAR:
         return false;
      case CLASS:
         return (arg_[0].class_.count() < REGEXP_CHMAX);
      case ALT:
         for (const RegExpr& r: arg_[0].arg_)
            if (r.op_ == STAR || !r.isCanonical()) 
               return false;
         return true;

      default: 
         return true;
      }
      break;

   case SEQ:
      if (!arg_[0].isCanonical() || !arg_[1].isCanonical()) return false;
      if (arg_[0].op_ == FALSE || arg_[1].op_ == FALSE) return false;
      if (arg_[0].op_ == EMPTY || arg_[1].op_ == EMPTY) return false;
      if (arg_[0].op_ == TRUE && arg_[1].inclEpsilon()) return false;
      return true;

   case ALT:
      for (const RegExpr& r: arg_)
         if (r.op_ == FALSE || r.op_ == TRUE || !r.isCanonical())
            return false;
      // Instance checks are more awkward to implement, so we skip them here.
      return true;

   default: 
      return false;
   }
}

bool RegExpr::
operator==(const RegExpr& re) const {
   if (op_ != re.op_) return false;
   if (op_ == CLASS) return (class_ == re.class_);
   if (op_ >= STAR) {
      if (arg_.size() != re.arg_.size())
         return false;
      else for (unsigned j=0; j < arg_.size(); j++)
         if (arg_[j] != re.arg_[j])
            return false;
   }
   return true;
}

bool RegExpr::
operator<(const RegExpr& re) const {
#define ltchk(a, b) if (a != b) return (a < b)
#define ltc(a) ltchk(a, re. a)
   ltc(op_);
   if (op_ == CLASS) {
      unsigned c = class_.count(); unsigned rc = re.class_.count();
      if (c != rc) return (c < rc);
      ltc(classData_[1]);
      ltc(classData_[0]);
      ltc(classData_[2]);
      ltc(classData_[3]);
      return false;
   }
   else if (op_ >= STAR) {
      ltc(arg_.size());
      for (unsigned j=0; j < arg_.size(); j++)
         ltc(arg_[j]);
   }

   return false;
}

size_t RegExpr::
hash() const {
   if (op_ == CLASS) {
      return fasthash64(classData_, sizeof(classData_), op_);
   }
   else {
      unsigned char buf[8 + (op_ >= STAR? 8*arg_.size() : 0)];
      unsigned j=0;
      size_t h;
      buf[j++] = op_;
      if (op_ >= STAR) {
         for (const RegExpr& re: arg_) {
            h = re.hash();
            memcpy(&buf[j], &h, 8); j+= 8;
         }
      }
      return fasthash64(buf, j, j);
   }
}

size_t RegExpr::
size() const {
   size_t rv= (op_ == SEQ? 0 : 1);
   if (op_ == SEQ || op_ == STAR || op_ == ALT)
      for (const auto& a: arg_)
         rv += a.size();
   return rv;
}

bool RegExpr::
isInstance(const RegExpr& re) const {
   if (op_ == FALSE || re.op_ == TRUE ||
       (isEpsilon() && re.inclEpsilon()))
      return true;
   else if (op_ == CLASS && re.op_ == CLASS) {
      bitset<REGEXP_CHMAX> res = class_ & ~re.class_;
      return (res.count() == 0);
   }
   else if (*this == re) 
      return true;
   else if (re.op_ == STAR) {
      if (isInstance(re.arg_[0]))
          return true;
      else 
         return (op_ == SEQ && arg_[0].isInstance(re) && arg_[1].isInstance(re));
   }
   else if (re.op_ == ALT) 
      for (const RegExpr& r: re.arg_)
         if (isInstance(r)) return true;
   return false;

   // @@@@ We can do more, e.g., 
   // x | y isInst u if x isInst u and y isInst u
   // x.y isInst u.z if z isInst u and y isInst z
   //
 }

void RegExpr::
deepCopy(const RegExpr& re) {
   assert(re.isCanonical());
   if (op_ >= STAR)
      arg_.clear();
   else if (re.op_ >= STAR) 
      new (&arg_) vector<RegExpr>();

   op_ = re.op_;
   if (op_ == CLASS)
      class_ = re.class_;
   else if (op_ >= STAR)
      arg_ = re.arg_;
}

#include <ctype.h>

void prtchar(int c, ostream& os) {
   if (isprint(c)) {
      if (c=='^' || c=='$') os << "\\";
      os << (char)c;
   }
   else os<< "\\0" << (char)(((c&0700)>>6)+'0') << (char)(((c&070)>>3)+'0') 
          << (char)((c&07)+'0');
}

void prtCharClass(const bitset<REGEXP_CHMAX>& cl1, ostream& os) {
   bitset<REGEXP_CHMAX> cl = cl1;
   bool needbrack = (cl.count() != 1);
   if (!needbrack) {
      if (cl.test(0)) {
         assert(cl.count() == 1);
         os << "$";
         return;
      }
   }
   if (needbrack)
      os << "[";
   if (cl.count() > 128) {
      os << "^";
      cl = ~cl;
   }
   if (cl.test('-') && (!cl.test('-'-1) || !cl.test('-'+1))) {
      os << '-';
      cl.reset('-');
   }
   for (int j=0; j<REGEXP_CHMAX; j++) {
      if (cl.test(j)) {
         prtchar(j, os);
         if (j < REGEXP_CHMAX-2 && cl.test(j+1) && cl.test(j+2)) {
            os << '-';
            for (j++; j < REGEXP_CHMAX; j++)
               if (!cl.test(j)) {
                  j--; break;
               }
            prtchar(j, os);
         }
      }
   }
   if (needbrack)
      os << "]";
}

void RegExpr::
print(ostream& os) const {
   switch (op_) {
   case FALSE: os << "_FALSE_"; break;
   case TRUE: os << "_TRUE_"; break;
   case EMPTY: os << "_EPSILON_"; break;
   case BEG: os << "^"; break;
   case CLASS: prtCharClass(class_, os); break;
   case STAR: os << '(' << arg_[0] << ")*"; break;
   case SEQ: os << arg_[0] << arg_[1]; break;
   case ALT: {
      bool first = true;
      os << '(';
      for (const RegExpr& re: arg_) {
         if (!first)
            os << '|';
         else first = false;
         os << re;
      }
      os << ')'; 
      break;
   }
   default: break; 
   }
}

#include <sstream>

string RegExpr::
getPrtString() const {
   stringstream s;
   print(s);
   return s.str();
}

bool RegExpr::
isString() const {
   switch (op_) {
   case CLASS: return (class_.count() == 1);
   case SEQ: return (arg_[0].isString() && arg_[1].isString());
   default:
      return false;
   }   
}

bool RegExpr::
inclEpsilon() const {
   switch (op_) {
   case TRUE:
   case EMPTY:
   case STAR:
      return true;
   case END:
      assert(false);
   case SEQ:
      return (arg_[0].inclEpsilon() && arg_[1].inclEpsilon());
   case ALT:
      for (const RegExpr& re: arg_)
         if (re.inclEpsilon())
            return true;
      return false;
   default:
      return false;
   }
}

bool RegExpr::
isEpsilon() const {
   switch (op_) {
   case EMPTY:
      return true;

   case CLASS:
      return (class_.count() == 0);

   case STAR:
   case SEQ:
   case ALT:
      for (const RegExpr& re: arg_)
         if (!re.isEpsilon())
            return false;
      return true;

   default:
      return false;
   }
}

void RegExpr::
residue(unsigned char c) {
   static unsigned char cl;
   if (c > cl) {
#if CG_DBG >= 2
   cout << *this << ".residue('" << c << "') clast=";
#endif
   residue(c, cl);
#if CG_DBG >= 2
   cout << cl << endl;
#endif
   }
   else residue(c, cl);
}

void RegExpr::
residue(unsigned char c, unsigned char& clast) {
   clast = REGEXP_CHMAX-1;
   switch (op_) {
   case FALSE:
   case EMPTY:
      op_ = FALSE;
      break;

   case TRUE:
      if (c == 0) {
         op_ = FALSE;
         clast = 0;
      }
      else op_ = TRUE;
      break;

   case END:
      assert(false);

   case CLASS:
      if (class_.test(c)) {
         op_ = EMPTY;
         for (unsigned char j = c; j < REGEXP_CHMAX; j++)
            if (class_.test(j)) 
               clast = j;
            else break;
      }
      else {
         op_ = FALSE; 
         for (unsigned char j = c; j < REGEXP_CHMAX; j++)
            if (!class_.test(j)) 
               clast = j;
            else break;
      }
      break;

   case STAR: {
      RegExpr re1(*this);
      arg_[0].residue(c, clast);
      RegExpr re2(SEQ, &arg_[0], &re1);
      deepCopy(re2);
      break;
   }

   case SEQ: {
      RegExpr re1 = arg_[0]; 
      RegExpr re2 = arg_[1];
      bool incEps = re1.inclEpsilon();
      re1.residue(c, clast);
      RegExpr nre(SEQ, &re1, &re2);
      if (incEps) {
         unsigned char c1;
         re2.residue(c, c1);
         nre = RegExpr(ALT, &nre, &re2);
         clast = min(clast, c1);
      }
      deepCopy(nre);
      break;
   }

   case ALT: {
      vector<RegExpr> vre(arg_);
      arg_.clear();
      for (unsigned j=0; j < vre.size(); j++) {
         unsigned char c1;
         vre[j].residue(c, c1);
         clast = min(clast, c1);
      }
      *this = vre[0];
      for (unsigned j=1; j < vre.size(); j++)
         merge(vre[j]);
      break;
   }

   default:
      assert(false);
   }

   if (c == 0)
      assert(op_ == TRUE || op_ == FALSE || op_ == EMPTY);
   else assert(isCanonical());

   //clast=c;
}

bitset<REGEXP_CHMAX> RegExpr::
first() const {
   bitset<REGEXP_CHMAX> rv;
   switch (op_) {
   case EMPTY:
      return rv; // empty set

   case TRUE:
      rv.set(0);
      rv.flip();
      return rv;

   case END:
      assert(false);

   case CLASS:
      return class_;

   case STAR:
      return arg_[0].first();

   case SEQ:
      rv = arg_[0].first();
      if (arg_[0].inclEpsilon())
         rv |= arg_[1].first();
      return rv;

   case ALT:
      for (const RegExpr& r: arg_)
         rv |= r.first();
      return rv;

   default:
      assert(false);
      //return rv; 
      // return empty set. This is questionable for FALSE,
      // on which first() should never be invoked.
   }
}
