#ifndef REG_EXPR_H
#define REG_EXPR_H

#include <bitset>
#include <iostream>
#include <assert.h>
#include <vector>

#include "Error.h"

using namespace std;

#define REGEXP_CHMAX 128

class CharClass {
   bool isNeg_;
   char rangeFirstCh_;
   string chars_;
 public:
   CharClass(bool neg=false): isNeg_(neg), rangeFirstCh_(0), chars_("") {};
   void addChar(char c) {
      if (rangeFirstCh_ != '\0') {
         if ((unsigned char)rangeFirstCh_+1 >= (unsigned char)c) {
             errMsg("Invalid range in regular expression");
         }
         else {
            for (unsigned j=(unsigned char)(rangeFirstCh_+1); 
                   j <= (unsigned char)c; j++)
               chars_ += (char)j;
         }
         rangeFirstCh_ = '\0';
         return;
      }
      if (c == '-' && chars_.size() > 0 && rangeFirstCh_ == '\0') {
         rangeFirstCh_ = chars_[chars_.size()-1];
      }
      else chars_ += c; 
   };

   bool negated() const {return isNeg_; };
   const string& chars() const { return chars_; };
};

class RegExpr;
ostream& operator<<(ostream& os, const RegExpr& re);

class RegExpr {
 public:
   enum RegExpOp: char 
     {TRUE, FALSE, EMPTY, BEG, END, CLASS, STAR, PLUS, QMARK, SEQ, ALT};
   // @@@@ For many, many regexpr: we shd add a STR op here so that plain strings
   // @@@@ can be stored in n bytes. Current rep needs n nodes of ~40B each.

   // Add addl operators such BEGIN_MATCH, END_MATCH to remember text,
   // and to support middle-of-pattern actions.

 private:
   RegExpOp op_;
   vector<RegExpr> arg_;
   union {
      bitset<REGEXP_CHMAX> class_;
      uint64_t classData_[REGEXP_CHMAX/(8*sizeof(uint64_t))];
   };

 public:
   RegExpr(): op_(EMPTY) {};
   RegExpr(char c): op_(CLASS), class_() {class_.set(c);};
   RegExpr(const CharClass& cc);
   RegExpr(const string& s);
   RegExpr(RegExpOp op, const RegExpr* arg1=nullptr, const RegExpr* arg2=0);
   RegExpr(const RegExpr& re) { op_ = EMPTY; deepCopy(re); };
   ~RegExpr() {if (op_ >= STAR) arg_.clear(); }

   bool isPhi() const { return op_ == FALSE; };
   bool isTrue() const { return op_ == TRUE; };
   bool isEpsilon() const;
   bool inclEpsilon() const;
   bool isCanonical() const;

   size_t size() const;

   bool isString() const;
   string getPrtString() const; // returns printed representation as string

   bool isInstance(const RegExpr& re) const;
   void merge(const RegExpr& re);
   void residue(unsigned char c);
   void residue(unsigned char c, unsigned char& clast);
   bitset<REGEXP_CHMAX> first() const;

   void print(ostream& os) const;
   void deepCopy(const RegExpr& re);

   const RegExpr& operator=(const RegExpr& re) { deepCopy(re); return *this; };

   bool operator==(const RegExpr& re) const;
   bool operator!=(const RegExpr& re) const { return !(*this == re); };
   bool operator<(const RegExpr& re) const;
   size_t hash() const;
};

inline ostream& operator<<(ostream& os, const RegExpr& re) {
   re.print(os);
   return os;
}

namespace std {
template<> 
struct hash<RegExpr> {
   size_t operator()(const RegExpr& re) { return re.hash(); };
};
};
#endif
