#include "STEClasses.h"
#include "ParserUtil.h"
#include "Ast.h"
#include "CondFac.h"
#include <assert.h>
#include "ST.h"

vector<SymTab*> SymTabEntry::symTabs_;

static const char* const kindtag[] = {
  "UNKNOWN_KIND",
  "GLOBAL_KIND","INTERFACE_KIND","STRUCT_KIND", "CLASS_KIND","FUNCTION_KIND", 
  "MODULE_KIND", 
  "VARIABLE_KIND",
  "ENUM_KIND", "ENUM_ELEM_KIND", 
  "EVENT_KIND", 
  "LABEL_KIND", "ON_KIND",
  "BLOCK_KIND", "RULE_BLOCK_KIND", "EVENT_BLOCK_KIND",
  "ERROR_KIND"
};

string kindTag(SymKind k) {
  if ((k >= 0) && (k < ERROR_KIND))
    return kindtag[k];
  else return kindtag[ERROR_KIND];
}

SymTabEntry::SymTabEntry(string name, SymKind k, int flags,
			 int lineno, int columnno, string filename, const Type *t) 
  : ProgramElem(t, flags, lineno, columnno, filename) {
  name_ = name;
  kind_ = k;
  if (type() == nullptr)
    type(new Type(Type::voidType));
  st_ = nullptr;
  next_ = prev_ = nullptr;
}

bool 
SymTabEntry::operator==(const SymTabEntry& ste)const {
  // @@@@ Reference comparison implemented, but is this appropriate?
  // Probably: STE cannot be copied or assigned, and there isn't much
  // reason to create multiple STEs for the same symbol, so this is enough
  return (&ste == this);
}

void 
SymTabEntry::usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
  SymTab *st = symTab();
  if (st != nullptr) 
    for (SymTab::iterator it = st->begin(); (it != st->end()); ++it)
      (*it)->usedSyms(use, assign);
}

const Type*
SymTabEntry::typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes) {
  SymTab *st = symTab();
  if (st != nullptr) 
    for (SymTab::iterator it = st->begin(); (it != st->end()); ++it) {
      SymTabEntry* ste = (*it);
      if (ste->kind() != BLOCK_KIND)
         ste->typeCheck(avlParams, valVtypes);
    }
  return &(Type::voidType);
}

ExprNode *
SymTabEntry::storageAlloc() {
  // This works for entries that are collections -- simply propagates
  // storageAlloc to components, and adds up and returns the result 
  SymTab *st = symTab();
  ExprNode *rv = nullptr;
  if (st != nullptr) {
    for (SymTab::iterator it = st->begin(); (it != st->end()); ++it) {
      ExprNode *s = (*it)->storageAlloc();
      if (s != nullptr) {
         if (rv == nullptr)
            rv = s;
         else rv = new OpNode(OpNode::PLUS, rv, s);
      }
    }
  }
  return rv;
}

const Type* 
GlobalEntry::typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes) {
   // First, perform type-checking on all children (i.e., the whole program)
   SymTabEntry::typeCheck(avlParams, valVtypes);

   // Next, walk through *ALL* the symbol tables in the specification, and
   // check declared-without-use and declared-without-assign

   set<SymTabEntry*> used;
   set<SymTabEntry*> assigned;
   usedSyms(used, assigned);

   for (auto& st: SymTabEntry::symTabs_) {
      // Don't complain about unused enum or struct elements; or unused function
      // parameters when there is no function body.
      if (st->parent() && 
          (st->parent()->kind() == ENUM_KIND   ||
           st->parent()->kind() == STRUCT_KIND ||
           (st->parent()->kind() == FUNCTION_KIND &&
            ((FunctionEntry*)st->parent())->body() == nullptr)))
         continue;

      for (SymTab::iterator it = st->begin(); (it != st->end()); ++it) {
         SymTabEntry* ste = (*it);
         // Don't complain about unused interface declarations or the main module
         if ((ste->flags() & extSym) ||
             (ste->kind() == MODULE_KIND && ste->name() == "main")) continue;
         if (ste->kind() == VARIABLE_KIND) {
            VariableEntry *ve = (VariableEntry*)ste;
            VarKind vk = ve->varKind();
            // Unused event params are common, plus they are used over such
            // short span of text that the warning is not even necessary.
            if (vk == EVENT_PARAM_VAR) continue;
            if (assigned.find(ve) == assigned.end()) {
               if (used.find(ve) == used.end())
                  warnMsg(ve->name() + " declared but never used", ve);
               else if (vk != FOREACH_VAR && vk != FN_PARAM_VAR && 
                        vk != MODULE_PARAM_VAR)
                  warnMsg(ve->name() + " used but never assigned", ve);
            }
            else if (used.find(ve) == used.end() && vk != FN_PARAM_VAR)
               warnMsg(ve->name() + " assigned but never used", ve);
         }
         /* Often, we define functions and modules in "libraries" and reuse
            them. It is natural for some of the symbols to be unused in such
            cases, so we might as well not complain. 
         else if (used.find(ste) == used.end() && ste->name()[0] != '_' &&
                  (ste->kind() == STRUCT_KIND || ste->kind() == FUNCTION_KIND ||
                   ste->kind() == MODULE_KIND))
            // Events should always occur within interfaces, so there is no need
            // to complain about unused events. To detect unused enum, we have
            // to check that the enum *and* its elements are never used. This
            // is a bit complex, with unclear benefits.
            warnMsg(ste->name() + " declared but never used", ste);
         */
      }
   }
   return new Type(Type::voidType);
}

void GlobalEntry::
codeGen(ostream& os, int indent, CGFlags& flags) {
   string ifcnm;
   ModuleEntry* mainme=nullptr;
   FunctionEntry* mainfun=nullptr;
   if (symTab() == nullptr) {
      errMsg("Empty input");
      return;
   }
   os << "#define ESTAR\n";
   for (auto ste: *symTab()) {
      CGFlags nfl(flags);
      if (ste->kind() == INTERFACE_KIND) {
         ifcnm = ste->name();
         os << "#include \"" << ifcnm << "_pre.h\"\n\n";
      }
      else if ((ste->kind() == FUNCTION_KIND || ste->kind() == ENUM_KIND) &&
               (ste->flags() & extSym) == 0) {
         if (ste->name() == "main")
            mainfun = (FunctionEntry*)ste;
         ste->codeGen(os, indent, nfl);
         endln(os, indent);
      }
      else if (ste->kind() == MODULE_KIND) {
         ((ModuleEntry*)ste)->codeGen(os, indent, nfl);
         if (ste->name() == "main") 
            mainme = (ModuleEntry*)ste;
      }
   }

   if (mainfun && !mainme)
      return;
   // @@@@ If there is just a main function, it does not work exactly the way we
   // @@@@ want: some needed header files are not included. This should be fixed,
   // @@@@ instead of relying on some fake interface definition which ends up
   // @@@@ including those header files.
   else if (mainme && mainfun) {
      errMsg("You cannot have both a main module and a main function");
      return;
   }
   if (!mainme) {
      errMsg("No main function or main module defined, aborting");
      return;
   }

   indent += STEP_INDENT; prtln(os, indent);

   // Now, instantiate all the modules. This will result in the generation of
   // a struct for each module, and a function corresponding to the reaction
   // of each rule. In addition, we will collect a global list of all rules
   // from all instantiated modules. (Remember, there will be a new copy of rules
   // for each instantiation.) 

   CGFlags nfl(flags);
   auto rs = mainme->instantiate(vector<ExprNode*>(), nfl);

   // Now, generate a function for each possible event that matches all of the
   // collected rules, and dispatches a call to each rule that matched.

   // Iterate through each possible event ...
   GlobalEntry* ge = (GlobalEntry*)stm.currentScope(GLOBAL_KIND);
   for (auto ste: *ge->symTab())
     if (ste->kind() == EVENT_KIND)
       evCodeGen(rs, (EventEntry*) ste, os, indent, nfl);

   prtRewind(os, STEP_INDENT, indent);
   os << "}";
   endln(os, indent);
   if (ifcnm != "")
      os << "\n#include \"" << ifcnm << "_post.h\"\n";
}

void 
GlobalEntry::print(ostream& os, int indent) const {
  emitST(os, indent-STEP_INDENT, '\0', '\0', true);
}

/****************************************************************/

void 
BlockEntry::print(ostream& os, int indent) const {
  emitST(os, indent, '{', '}', false);
}

/****************************************************************/

void 
ClassEntry::print(ostream& os, int indent) const {
  os << "class " << name() << ';';
}

/****************************************************************/

void 
FunctionEntry::usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   SymTabEntry::usedSyms(use, assign);
   if (body_)
      body_->usedSyms(use, assign);
}


const Type*
FunctionEntry::typeCheck(set<const SymTabEntry*>& avlParamsIgnored, 
                        map<VarKind, bool>& valVtypesIgnored) {
   set<const SymTabEntry*> avlParams;
   map<VarKind, bool> valVtypes = {
      {FOREACH_VAR, false}, {FN_PARAM_VAR, true}, {LOCAL_VAR, true},
      {MODULE_PARAM_VAR, false}, {EVENT_PARAM_VAR, false}, 
      {GLOBAL_VAR, !isConst()}};

   SymTabEntry::typeCheck(avlParams, valVtypes);

  if (body_ != nullptr)
     body_->typeCheck(avlParams, valVtypes);
  return type();
}

void FunctionEntry::codeGen(ostream& os, int indent, CGFlags& flags) {
  /* The easiest way to implement functions is to translate
     a function defn into an equivalent function in the target
     language. Does this constrain the types of arguments to
     a function? It should not -- otherwise, it wont be useful to
     define BMSL functions, but instead use functions 
     defined in the target language. */

    if (name() != "main" && 
         body_ != nullptr) // if body exists, fn will appear within a module,
       os << "static ";    // i.e., within class scope, so make it static.
    os << type()->retType()->fullCodeName();
    os << " " << flags.prefix() << name();
    if (type() != nullptr) {
       if (type()->arity() > 0)
          // Emit parameter declarations
          emitST(os, indent, '(', ')', false, &flags, 0, type()->arity());
       else os << "()";
       if (body_ != nullptr) {
          os << " {"; 
          indent += STEP_INDENT; prtln(os, indent);
          if (safeCoercions) {
             os << "tpvardecls"; endln(os, indent);
          }
          // Emit remaining (local variable) declarations
          emitST(os, indent, '\0', ';', true, &flags, type()->arity(), 10000);
          // Emit the body code.
          body_->emitWithoutBraces(os, indent, flags);
          indent -= STEP_INDENT;
          prtln(os, indent);
          os << '}';
       }
    }
}

void 
FunctionEntry::print(ostream& os, int indent) const {
  if ((type() != nullptr) && (type()->retType() != nullptr))
    type()->retType()->print(os);
  
  os << " " << name();
  if (type() != nullptr) {
	if (type()->arity() > 0)
	  emitST(os, indent, '(', ')', false, nullptr, 0, type()->arity());
	else os << "()";
	if (body_ != nullptr) {
	  os << " {"; prtln(os, indent+STEP_INDENT);
	  emitST(os, indent, '\0', ';', true, nullptr, type()->arity(), 10000);
	  prtSpace(os, STEP_INDENT);
      CGFlags fl("_NONE_");
	  body_->emitWithoutBraces(os, indent, fl);
      prtln(os, indent);
	  os << '}';
	}
	else os << ';';
  }
  else {
	emitST(os, indent, '(', ')', false);
	if (body_ != nullptr)
	  body_->print(os, indent);
	else os << ';';
  }
}

/****************************************************************/

void 
EventEntry::codeGen(ostream& os, int indent, CGFlags& flags) {
  if (flags.defn_ && !flags.invocation_) 
     os << "ET_EVENT_RET ";
  os << name();
  if (!flags.defn_ && !flags.invocation_) {
     emitST(os, indent, '(', ')', false);
     os << ';';
  }
  else {
     unsigned i=0;
     os << '(';
     if (symTab() != nullptr)
        for (auto it = symTab()->begin(); it != symTab()->end(); ++it, ++i) {
           assert((*it)->kind() == VARIABLE_KIND);
           VariableEntry* ve = (VariableEntry*)(*it);
           assert(ve->varKind() == EVENT_PARAM_VAR);
           if (i != 0) os << ", ";
           if (!flags.invocation_) {
              if (!Type::isPrimitive(ve->type()->tag()))
                 os << "const ";
              os << ve->type()->fullCodeName(true);
              if (!Type::isPrimitive(ve->type()->tag()))
                 os << '&';
           }
           os << " __p" << i;
        }
     os << ')';
  }
}

void 
EventEntry::print(ostream& os, int indent) const {
  os << "event "<< name();
  emitST(os, indent, '(', ')', false);
  os << ';';
}

/****************************************************************/

/*
VariableEntry::VariableEntry(const VariableEntry &v):
  SymTabEntry(v.name(), VARIABLE_KIND, v.flags(), v.line(), 
			  v.column(), v.file(), new Type(*v.type())) {
  vkind_ = v.vkind_;
  if (v.offset_ != nullptr)
	offset_ = (ExprNode*)v.offset_->clone();
  else offset_ = nullptr;
  if (v.initVal_ != nullptr) 
	initVal_ = (ExprNode*)v.initVal_->clone();
  else initVal_ = nullptr;

  for (unsigned int i=0; i < v.subscripts_.size(); i++)
      subscripts_.push_back((ExprNode*)v.subscripts_[i]->clone());
}
*/

void 
VariableEntry::usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (initVal_ != nullptr) {
      assign.insert(this);
      initVal_->usedSyms(use, assign);
   }
   for (auto& subs: subscripts_)
      subs->usedSyms(use, assign);
}

const Type* 
VariableEntry::typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes) {
   assert(/*initVal_ == nullptr &&*/ subscripts_.size() == 0);
   // @@@@ initVal_ == null is there *perhaps* because of some difficulty with 
   // @@@@ providing decl-without-use and use-without-assign warnings.
   // @@@@ Until we figure it out, we accept those problems.

  if (offset_) {
     //set<const SymTabEntry*> avlParams;
     //map<VarKind, bool> valVt = {{MODULE_PARAM_VAR, false}};
     offset_->typeCheck(avlParams, valVtypes);
     offset_->coerce(&Type::intType);
  }

  if (type() == nullptr) {
     if (varKind() == RULE_VAR)
        type(new Type(Type::NATIVE));
     else errMsg("Missing type for variable " + name(), this);
  }

  if (initVal_) {
     RefExprNode *lhs = new RefExprNode(nullptr, name(), this, -1, 
                        initVal_->line(), initVal_->column(), initVal_->file());
     OpNode* n = new OpNode(OpNode::ASSIGN, lhs, initVal_, -1, 
                        initVal_->line(), initVal_->column(), initVal_->file());
     n->typeCheck(avlParams, valVtypes);
  }
  return type();
}

/*
const Type* 
VariableEntry::assignChk(ExprNode* rhs, set<const SymTabEntry*>& avlPar, 
                         map<VarKind, bool>& valVtypes, const ProgramElem *pe) {
   if (!pe) pe = this;
   RefExprNode *lhs = new RefExprNode(nullptr, name(), this, -1, pe->line(), 
                                      pe->column(), pe->file());
   OpNode* n = new OpNode(OpNode::ASSIGN, lhs, rhs, -1, pe->line(), 
                                      pe->column(), pe->file());
   n->typeCheck(avlPar, valVtypes);
/ *********
   Type* lhst = type();
   if (lhst && lhst->isValid()) {
      rhs->typeCheck(avlPar, valVtypes);
      const Type* rhst = rhs->type();
      if (rhst && rhst->isValid()) {
         if (valVtypes.find(varKind()) == valVtypes.end() || 
             !valVtypes[varKind()]) {
            errM2("Variable " << name() << " cannot be updated in this " <<
                  "context", pe? pe : this);
         }
         else if (varKind() == RULE_VAR) {
            // A rule variable's type should be inferred from first assignment.
            type(new Type(*rhst));
         }
         else if (!rhst->isSubType(*lhst)) {
            errM2("Assigned type '" << *rhst << "' not a subtype of target "
                  << "type '" << *lhst << "'", pe? pe : this);
         }
         else if (!lhst->isSubType(*rhst))
            rhs->coerce(new Type(*lhst));
      }
   }
   if (type()->isConst())
      errM2("const variable " << name() << " cannot be updated", pe? pe : this);
**** /
   return type();
}
*/

/*
const Type* VariableEntry::
typeCheckInit(set<const SymTabEntry*>& avlPar, map<VarKind, bool>& valVtypes) {
  const Type *rv;
  if (initVal_ != nullptr) 
     assignChk(initVal_, avlPar, valVtypes);

  if (offset_ != nullptr) {
     rv = offset_->typeCheck(avlPar, valVtypes);
     if (rv == nullptr || !Type::isIntegral(rv->tag())) {
        internalErr("Offset must be of integral type");
        return type();
     }
  }
  //
  //for (unsigned i=0; i < subscripts_.size(); i++) {
  //   rv = subscripts_[i]->typeCheck(avlPar, valVtypes);
  //   if (rv == nullptr || !Type::isIntegral(rv->tag())) {
  //      errMsg("Subscript " + itoa(i) + " must be of integral type", this);
  //      return type();
  //   }
  //}
  //
  // @@@@ There is some complexity introduced by rule variables ---
  // we cannot get their type until we typecheck the expression that is
  // assigned to them. But expression typechecking follows that of entries. 
  // Our approach: when typechecking assignments, set the type of lhs
  // if it is not already set.

  if (type() == nullptr) {
     if (varKind() == RULE_VAR)
        type(new Type(Type::NATIVE));
     else errMsg("Missing type for variable " + name(), this);
  }
  return type();
}
*/

void 
VariableEntry::codeGen(ostream& os, int indent, CGFlags& flags) {
  if (type() != nullptr)
    os << type()->fullCodeName(true);

  if ((vkind_ == FN_PARAM_VAR || vkind_ == EVENT_PARAM_VAR) 
      && !Type::isPrimitive(type()->tag()))
     os << '&';
  
  os << ' ' << flags.prefix() << name();  
  for (unsigned int i = 0; i < nSubscripts(); i++) {
	os << '[';
	subscript(i)->codeGen(os, indent, flags);
	os << ']';
  }
  
  if (initVal_) {
    os << " = ";
    initVal_->codeGen(os, indent, flags);
  }
  else if ((vkind_ == LOCAL_VAR || vkind_ == GLOBAL_VAR) && 
           (type()->isNumeric() || type()->isBool())) {
     os << " = ";
     Value::zero().codeGen(os, type(), flags.prefix());
  }
}

void 
VariableEntry::print(ostream& os, int indent) const {
  if (type() != nullptr)
    type()->print(os);
  
  os << ' ' << name();  
  for (unsigned int i = 0; i < nSubscripts(); i++) {
	os << '[';
	subscript(i)->print(os, indent);
	os << ']';
  }
  
  if (initVal_ != nullptr) {
    os << " = ";
    initVal_->print(os,indent);
  }
}

/****************************************************************/

void 
EnumEntry::codeGen(ostream& os, int indent, CGFlags& flags) {
   os << "enum " << flags.prefix() << name();
   if (elems_ != nullptr) {
	 os << " {";
	 for (unsigned i=0; i < elems_->size(); i++) {
	   if ((*elems_)[i] != nullptr) {
		 os << flags.prefix() << (*elems_)[i]->name();
		 if (i < elems_->size()-1)
		   os << ", ";
	   }
	 }
	 os << "}";
  }
}

void 
EnumEntry::print(ostream& os, int indent) const {
  if (name() != "states")
    os << "enum ";
  os << name();
  if (elems_ != nullptr) {
	 os << " {";
	 for (unsigned i=0; i < elems_->size(); i++) {
	   if ((*elems_)[i] != nullptr) {
		 os << (*elems_)[i]->name();
		 if (i < elems_->size()-1)
		   os << ", ";
	   }
	 }
	 os << "};";
  }
}

/****************************************************************/

StructEntry::StructEntry(string name, int flags,int line,int column,string file):
  SymTabEntry(name, STRUCT_KIND, flags, line, column, file) {
  type(new Type(this, Type::STRUCT));
  size_ = nullptr;
}

void 
StructEntry::codeGen(ostream& os, int indent, CGFlags& flags) {
   assert(0);
}

void 
StructEntry::print(ostream& os, int indent) const {
  os << "struct " << name();
  if (parent_.size() > 0)
    os << ": ";
  unsigned int i = 0;
  while ((i < parent_.size()) && (i < constraint_.size())) {
    os << parent_[i]->name();
    if (constraint_[i] != nullptr) {
      os << " with (";
      constraint_[i]->print(os, indent);
      os << ')';
    }
    if (++i < parent_.size()) {
      os << " or ";
      prtln(os, indent+STEP_INDENT);
    }
  }
  
  os << "{"; prtln(os, indent+STEP_INDENT);
  emitST(os, indent, '\0', ';', true);
  os << '}';
}

const SymTabEntry*
StructEntry::lookUpWithInheritance(string var, const ProgramElem *pe) const {
  const SymTabEntry *rv = nullptr;
  if (symTab() != nullptr) {
    rv = symTab()->lookUp(var);
    if (rv != nullptr)
      return rv;
  }

  unsigned int npar=0;
  const SymTabEntry* t;
  const StructEntry *par;
  for (unsigned int i=0; i < parent_.size(); i++) {
    t = parent_[i]->lookUpWithInheritance(var, pe);
    if (t != nullptr) {
      rv = t;
      if (npar > 1) {
		errMsg(string("Error looking up name '" ) + var + "', which appears"
			   " in multiple base classes of Struct "+ name()+
			   ":  structs " + parent_[i]->name() + " and " + par->name(), pe);
		return nullptr;
      }
	  else par = parent_[i];    }
  }
  return rv;
}

/****************************************************************/

ModuleEntry::ModuleEntry(string name, int fl, int line, int column, string file):
  SymTabEntry(name,MODULE_KIND,fl, line, column, file, new Type(Type::voidType)),
  timeout_(), mapDecls_(), onDecls_(), 
  rules_() {
  isStateMachine_ = false;
  startState_ = finalState_ = nullptr;
}

void 
ModuleEntry::usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   SymTabEntry::usedSyms(use, assign);
   for (auto it = timeout_.begin(); it != timeout_.end(); ++it)
      it->second->usedSyms(use, assign);

   for (auto& it: mapDecls_)
      it->usedSyms(use, assign);

   for (auto& it: onDecls_)
      it->usedSyms(use, assign);

   for (auto& it: rules_)
      it->usedSyms(use, assign);

   for (auto& it: instantiate_)
      it->usedSyms(use, assign);
}

const Type*
ModuleEntry::typeCheck(set<const SymTabEntry*>& avlParamsIgnored, 
                        map<VarKind, bool>& valVtypesIgnored) {

  set<const SymTabEntry*> avlParams;
  SymTab *st = symTab();
  vector<VariableEntry*> modparams;

  // We are now allowing module parameters to be variables that are assigned
  if (st != nullptr) {
     map<VarKind, bool> valVtypes = {
        {FOREACH_VAR, false}, {LOCAL_VAR, true},
        {MODULE_PARAM_VAR, false}, {EVENT_PARAM_VAR, false}, 
        {GLOBAL_VAR, true}};

    for (SymTab::iterator it = st->begin(); (it != st->end()); ++it) {
      SymTabEntry* ste = (*it);
      if (ste->kind() != BLOCK_KIND) {
         const Type* t = ste->typeCheck(avlParams, valVtypes);
         if (ste->kind() == VARIABLE_KIND &&
             ((VariableEntry*)ste)->varKind() == MODULE_PARAM_VAR)
            modparams.push_back((VariableEntry*)ste);
      }
    }
  }
 
  if (isStateMachine_) {
     if (startState_ == nullptr)
        errMsg("Start state must be specified for state machine "+name(), this);
     else if (finalState_ == nullptr)
        errMsg("Final state must be specified for state machine "+name(), this);
     else if (startState_ == finalState_)
        errMsg("Start and final states can't be same in module "+ name(), this);

     map<VarKind, bool> valVtgr = { 
           {FOREACH_VAR, false}, {FN_PARAM_VAR, true}, {MODULE_PARAM_VAR, false},
           {EVENT_PARAM_VAR, false}, {GLOBAL_VAR, false}};
     map<VarKind, bool> valVtng = { 
           {FOREACH_VAR, false}, {FN_PARAM_VAR, true}, 
           {MODULE_PARAM_VAR, false}, {EVENT_PARAM_VAR, false}};
     for (auto it = timeout_.begin(); it != timeout_.end(); ++it) {
        if (!it->second->isConst()) {
           errMsg("Timeouts must be constant-valued", this);
           break;
        }
        if (!it->second->coercedType() && !it->second->type())
           it->second->typeCheck(avlParams, valVtng);
     }

     for (auto it = mapDecls_.begin(); it != mapDecls_.end(); ++it)
        if (!(*it)->typeCheck(avlParams, valVtgr)->isValid())
           break;
     // @@@@ Is there completenes criteria for map stmts? If so, check them.
  }
  else {
     if (startState_ != nullptr)
        errMsg("State machine constructs cannot be used in module " + name() +
               ", which does not specify states", startState_);
     else if (finalState_ != nullptr)
        errMsg("State machine constructs cannot be used in module " + name() +
               ", which does not specify states", finalState_);
     else if (timeout_.size() != 0)
        errMsg("Module " + name() + 
               ": Timeout supported only in state machines",  this);
     else if (mapDecls_.size() != 0)
        errMsg("Module "+ name() + ": Map statement supported only in state "
               "machine", mapDecls_.front());
  }

  // @@@@ Need a way to permit functions (and other modules?) to be passed as
  // @@@@ arguments. This probably requires generalizing the type system, or 
  // @@@@ at the least, revising InvocationNode::typeCheck

  map<VarKind, bool> valVts = {{MODULE_PARAM_VAR, false}};
  for (unsigned i=0; i < instantiate_.size(); i++)
     instantiate_[i]->typeCheck(avlParams, valVts);

  set<SymTabEntry*> used, assigned;
  usedSyms(used, assigned);
  for (VariableEntry* mpve: modparams) {
     Type* mpt = mpve->type();
     // Module param is not like a function param, so const is inapplicable
     if (mpt && mpt->isConst())
        warnMsg("'const' qualifier meaningless for module parameter type"
                ", ignoring ", this);
     // The real question of whether a variable or a constant should be 
     // passed in as a module parameter is determined by how this param is
     // is used in the module. If it is assigned, then the parameter should
     // be a variable (refexprnode). Otherwise, it can be a constant.

     //cout<<name() <<": "<< mpve->name() << ": "<< assigned.count(mpve) << endl;
     mpve->type()->constType(assigned.count(mpve) == 0);
  }

  map<VarKind, bool> valVtgw; // rule typeCheck ignores this, so don't initialize

  for (unsigned int i=0; i < rules_.size(); i++) {
     rules_[i]->typeCheck(avlParams, valVtgw);
  }

  return &(Type::voidType);
}

void 
ModuleEntry::codeGen(ostream& os, int indent, CGFlags& flags) {
  // Generate a struct (class) to for the module. This struct contains:
  //   1. A data member for (a) each state var, (b) each on-statistic, and
  //      (c) each instantiated module (a struct member)
  //   2. A function for each rule that contains the code for the rule's
  //      reaction, plus the associated on-statistic maintenance.
  //
  // All calls to these rule functions are made from the main module's code. The
  // main module contains one function for each event. This function contains
  // the code for matching all rules in all instantiated modules. Reaction
  // functions (generated in step 2) will then be called for each matching rule.

   CGFlags nfl(flags);
   os << "struct __" << name() << " {"; 
   indent += STEP_INDENT; prtln(os, indent);

   // @@@@ We are not able to use module parameters for variable initialization
   // @@@@ or within on declarations. To support this, we should introduce a
   // @@@@ state var _a for each module parameter a. By setting subst_[a] = _a,
   // @@@@ we should be able to replace all uses of a in on statements with _a.
   // @@@@ In addition, we shd emit _a(a). Plus, state variables should be
   // @@@@ initialized after the : in the constructor. 

   // Generate a data member for each state variable
   if (symTab() != nullptr) {
      for (auto ste: *symTab()) {
         if (ste->kind() == VARIABLE_KIND) {
            auto ve = (VariableEntry*)ste;
            if (ve->varKind() == GLOBAL_VAR) {
               ste->codeGen(os, indent, nfl);
               // @@@@ Initialization may need to be moved into constructor
               endln(os, indent);
            }
         }
      }
   }

   // Next, the data members for maintaining "on" statistics
   for (auto it: onDecls_)
      it->emitVars(os, indent, nfl);

   // Finally, (struct) data members for each instantiated module.
   int k=0;
   for (auto it: instantiate_) {
      InvocationNode& mod = *it; 
      ModuleEntry *me = (ModuleEntry*)(mod.symTabEntry());
      os << "__" << me->name() << " __ms" << k++;
      endln(os, indent);
   }

   // Onto function members. Start with the constructor
   os << "__" << name();
   int nargs =  (type() != nullptr) ? type()->arity() : 0;
   if (nargs > 0)
      emitST(os, indent, '(', ')', false, nullptr, 0, nargs);
   else os << "()";
   if (instantiate_.size() > 0 || onDecls_.size() > 0)
      os <<": ";
   nfl.first_ = true;
   int kk=0;
   for (auto it: instantiate_) {
      InvocationNode& mod = *it; 
      if (nfl.first_)
         nfl.first_ = false;
      else os << ", ";
      os << "__ms" << kk++ << "(";
      if (mod.params() != nullptr)
         for (unsigned j=0; j < mod.params()->size(); j++) {
            if (j > 0) os << ", ";
            (*mod.params())[j]->codeGen(os, indent, nfl);
         }
      os << ")";
   }

   for (auto it: onDecls_)
      it->emitInit(os, indent, nfl);

   os << "{}"; endln(os, indent); // Finished with the constructor.

   // Generate a function corresponding to each reaction
   for (auto rule: rules_) {
      nfl.first_ = true; nfl.curRule_ = rule;

      string rf_name = "__" + rule->label()->name();
      os << "void " <<  rf_name << "(";
      indent += STEP_INDENT;

      FunctionEntry* rfe = new FunctionEntry(rf_name);
      vector<ExprNode*>* rf_params = new vector<ExprNode*>(); 

      // Identify the variables used, make them formal params to rule function
      set<SymTabEntry*> used;
      rule->reaction()->usedSyms(used, used);
      for (auto sym: used) {
         if (sym->kind() != VARIABLE_KIND) continue;
         VariableEntry* ve = (VariableEntry*)sym;
         if (ve->varKind() != EVENT_PARAM_VAR && 
             ve->varKind() != MODULE_PARAM_VAR) continue;
         if (nfl.first_)
            nfl.first_ = false;
         else os << ", ";
         os << ve->type()->fullCodeName(true);
         if (ve->varKind() == MODULE_PARAM_VAR &&
             !ve->type()->isConst())
            os << "& ";
         os << ' ' << ve->name();
         rf_params->push_back((ExprNode*)
                                 new RefExprNode(nullptr, ve->name(), ve));
      }

      os << ") {"; prtln(os, indent);
      StmtNode* s = rule->reaction();
      if (s->stmtNodeKind() == StmtNode::COMPOUND)
         ((CompoundStmtNode*)s)->emitWithoutBraces(os, indent, nfl);
      else s->codeGen(os, indent, nfl);
      prtln(os, indent);

      // Find applicable on statements and generate code for them.
      nfl.curRule_ = rule;
      for (auto on: onDecls_)
         on->codeGen(os, indent, nfl);

      indent -= STEP_INDENT; //prtln(os, indent);
      prtRewind(os, STEP_INDENT, indent);
      os << "}"; 
      endln(os, indent);

      // Finally, replace action with a call to the generated rule function
      InvocationNode* rf_call = new InvocationNode(rfe,  rf_params);
      rule->reaction(new ExprStmtNode(rf_call));
   }

   if (name() != "main") {
      indent -= STEP_INDENT;
      prtRewind(os, STEP_INDENT, indent);
      os << "}"; endln(os, indent);
   }
}

vector<RuleNode*>
ModuleEntry::instantiate(vector<ExprNode*> params, CGFlags& flags) const {
   if (params.size() > 0) {
      unsigned j=0;
      for (auto it: *symTab()) {
         flags.subst_[it] = params[j];
         if (++j >= params.size()) break;
      }
   }

   vector<RuleNode*> irules;

   unsigned i=0;
   for (auto it: instantiate_) {
      CGFlags nfl(flags);
      InvocationNode& mod = *it; 
      const ModuleEntry *me = (const ModuleEntry*)(mod.symTabEntry());
      vector<ExprNode*> el;
      if (mod.params() != nullptr) {
        for (unsigned j=0; j < mod.params()->size(); j++)
           el.push_back((ExprNode*)(*mod.params())[j]->clone(&nfl));
      }

      nfl.addPrefix("__ms" + to_string(i++));
      auto rv = me->instantiate(el, nfl);
      irules.insert(irules.end(), rv.begin(), rv.end());
   }

   for (auto r: rules_) {
      vector<RuleNode*> nrs = r->split();
      for (auto nr: nrs)
         irules.push_back((RuleNode*)nr->clone(&flags));
   }

   return irules;
}

void 
ModuleEntry::print(ostream& os, int indent) const {
  int nargs =  (type() != nullptr) ? type()->arity() : 0;

  if (isStateMachine())
	os << "statemachine ";
  else os << "module ";
  os << name();
  if (nargs > 0)
    emitST(os, indent, '(', ')', false, nullptr, 0, nargs);
  else os << "() ";
  
  os << " {"; prtln(os, indent+STEP_INDENT);
  int j=0;
  const SymTab* st = symTab();
  if (st != nullptr) {
     for (auto ste: *st) {
        if (ste->kind() == ON_KIND)
           break;
        j++;
     }
     if (j > st->size()) j = st->size();
  }
  emitST(os, indent, '\0', ';', true, nullptr, nargs, j);
  prtSpace(os, STEP_INDENT);

  if (startState_) {
    os << "startstate " << startState_->name();
	endln(os, indent+STEP_INDENT);
  }
  
  if (finalState_) {
    os << "finalstate " << finalState_->name();
	endln(os, indent+STEP_INDENT);
  }
  
  for (unsigned i=0; ((i < imports_.size()) && (imports_[i] != nullptr)); i++) {
	os << "import " << imports_[i]->name();
	endln(os, indent+STEP_INDENT);
  }
  
  for (unsigned i=0; i < instantiate_.size(); i++) {
	os << "instantiate " << *instantiate_[i];
	endln(os, indent+STEP_INDENT);
  }
  
  if (mapDecls_.size() > 0) {
    vector<PrimitivePatNode*>::const_iterator it = mapDecls_.begin();
    while (it != mapDecls_.end()) {
      os << "map ";
      (*it)->print(os, indent+STEP_INDENT);
	  endln(os, indent+STEP_INDENT);
      ++it;
    }
  }

  if (timeout_.size() > 0) {
    map<EnumElemEntry*, ExprNode*>::const_iterator it = timeout_.begin(); 
    while (it != timeout_.end()) {
       os << "timeout ";
       it->second->print(os, indent);
       os << " in ";
       it->first->print(os, indent);
       endln(os, indent+STEP_INDENT);
       ++it;
    }
  }

  for (int i = 0, n = rules_.size(); (i < n); i++) 
    if (rules_[i] != nullptr)
      rules_[i]->print(os,indent+STEP_INDENT);

  prtln(os, indent);
  emitST(os, indent, '\0', ';', true, nullptr, j, 10000);

  prtRewind(os, STEP_INDENT, indent);
  os << '}';
}

/****************************************************************/

OnEntry::OnEntry(string name, vector<RuleNode*>* trans, int fl, 
				 int line, int column, string file):
       SymTabEntry(name,ON_KIND,fl,line,column,file,new Type(Type::voidType)) {
  separateStatsForEachRule_ = false;
  size_ = cond_ = maxprob_ = learnCond_ = detectCond_ = nullptr;
  minInterval_ = numScales_ = nullptr;
  value_ = nullptr;
  trans_ = trans;
  wrt_ = nullptr;
}

void 
OnEntry::usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign) {
   if (size_) size_->usedSyms(use, assign);
   if (cond_) cond_->usedSyms(use, assign);
   if (maxprob_) maxprob_->usedSyms(use, assign);
   if (learnCond_) learnCond_->usedSyms(use, assign);
   if (detectCond_) detectCond_->usedSyms(use, assign);

   if (minInterval_) minInterval_->usedSyms(use, assign);
   if (numScales_) numScales_->usedSyms(use, assign);

   if (value_) value_->usedSyms(use, assign);

   if (wrt_ != nullptr)
      for (auto& it: *wrt_)
         it->usedSyms(use, assign);
}

const Type*
OnEntry::typeCheck(set<const SymTabEntry*>& avlParamsIgnored, 
                        map<VarKind, bool>& valVtypesIgnored) {
   set<const SymTabEntry*> avlParams;
   map<VarKind, bool> valVtypes = {
      {FOREACH_VAR, false}, {FN_PARAM_VAR, false}, {GLOBAL_VAR, false},
      {MODULE_PARAM_VAR, false}, {EVENT_PARAM_VAR, false}};


   if (size_ != nullptr) {
      // @@@@ Need to check if size value is within reasonable bounds
      if (!size_->isConst())
         errMsg("On statement: size must be constant-valued", this);
      else {
         const Type* t = size_->typeCheck(avlParams, valVtypes);
         if (t == nullptr || !Type::isIntegral(t->tag()))
            errMsg("On statement: invalid type for size", this);
         else size_->coerce(new Type(Type::uintType)); // @@@@ long type?
      }
   }       

   if (cond_ != nullptr) {
      const Type* t = cond_->typeCheck(avlParams, valVtypes);
      if (t == nullptr || !Type::isBool(t->tag()))
         errMsg("On statement: condition should be boolean", this);
      else cond_->coerce(new Type(Type::boolType));
   }

   // @@@ Need to check if 0 < maxprob_ < 1.0
   if (maxprob_ != nullptr) {
      if (!maxprob_->isConst())
         errMsg("On statement: false alarm rate must be constant-valued", this);
      else {
         const Type* t = maxprob_->typeCheck(avlParams, valVtypes);
         if (t == nullptr || !Type::isDouble(t->tag()))
            errMsg("On statement: invalid type for false alarm rate", this);
         else maxprob_->coerce(new Type(Type::doubleType));
      }
   }

   if (learnCond_ != nullptr) {
      const Type* t = learnCond_->typeCheck(avlParams, valVtypes);
      if (t == nullptr || !Type::isBool(t->tag()))
         errMsg("On statement: condition should be boolean", this);
      else learnCond_->coerce(new Type(Type::boolType));
   }

   if (detectCond_ != nullptr) {
      const Type* t = detectCond_->typeCheck(avlParams, valVtypes);
      if (t == nullptr || !Type::isBool(t->tag()))
         errMsg("On statement: condition should be boolean", this);
      else detectCond_->coerce(new Type(Type::boolType));
   }

   if (minInterval_ != nullptr) {
      // @@@@ Need to check if time within reasonable bounds
      if (!minInterval_->isConst())
         errMsg("On statement: timescales must be constant-valued", this);
      else {
         const Type* t = minInterval_->typeCheck(avlParams, valVtypes);
         if (t == nullptr || !Type::isNumeric(t->tag()))
            errMsg("On statement: invalid type for timescale", this);
         //else minInterval_->coerce(new Type(Type::uintType));
      }
   }

   if (numScales_ != nullptr) {
      // @@@@ Need to check if time within reasonable bounds
      if (!numScales_->isConst())
         errMsg("On statement: number of scales must be constant-valued", this);
      else {
         const Type* t = numScales_->typeCheck(avlParams, valVtypes);
         if (t == nullptr || !Type::isNumeric(t->tag()))
            errMsg("On statement: invalid type for timescale", this);
      }
   }

   if (value_ != nullptr) {
      const Type* t = value_->typeCheck(avlParams, valVtypes);
      if (t == nullptr || !Type::isNumeric(t->tag()))
         errMsg("On statement: invalid type for value distribution", this);
   }

   if (wrt_ != nullptr)
      for (auto it: *wrt_)
         it->typeCheck(avlParams, valVtypes);

   return &(Type::voidType);
}

vector<string>
OnEntry::emitVars(ostream &os, int indent, CGFlags& flags) {
   unsigned i=0; vector<string> rv;
   for (auto r: *trans_) {
      if (flags.curRule_ != nullptr && flags.curRule_ != r) continue;
      if (i == 0 || separateStatsForEachRule_) {
         if (flags.curRule_ != nullptr) {
            if (flags.first_)
               flags.first_ = false;
            else os << ", "; 
         }
         if (wrt_ != nullptr && wrt_->size() > 0) {
            os << "MFUTable <";
            if (wrt_->size() > 1) os << "tuple<";
            unsigned j=0;
            for (auto e: *wrt_) {
               if (j++ != 0) os << ", ";
               os << e->type()->codeName();
            }
            if (wrt_->size() > 1) os << ">";
            os << ", ";
         }
         if (minInterval_ != nullptr) {
            os << "FreqDistrib<";
            numScales_->codeGen(os, indent, flags);
            os << ">";
            // %%%% minInterval is not used here, but should be used as
            // %%%% argument to inc(...), which should figure out the details
         }
         else os << "GeoHistogram";
         if (wrt_ != nullptr) os << ">";

         string vnm = name() + "_" + to_string(i);
         rv.push_back(vnm);
         if (flags.curRule_ == nullptr) {
            os << " " <<  vnm << ";";
            prtln(os, indent);
         }
         else os << "& " << vnm;
      }
      i++;
   }
   return rv;
}

void 
OnEntry::emitInit(ostream &os, int indent, CGFlags& flags) {
   for (unsigned i=0; i < trans_->size(); i++) {
      if (i == 0 || separateStatsForEachRule_) {
         if (!flags.first_) {
            os << ", ";
            prtln(os, indent);
         }
         flags.first_ = false;
         os << " " << name() << "_" << to_string(i) << "(";
         if (wrt_ != nullptr && wrt_->size() > 0)
            size_->codeGen(os, indent, flags);
         os << ")"; 
      }
   }
}

void 
OnEntry::codeGen(ostream &os, int indent, CGFlags& flags) {
   unsigned i=0;
   // @@@@ These operations need the current time. The easiest way to route 
   // @@@@ current time is to have the data structures read an extern variable.
   for (auto r: *trans_) {
      if (flags.curRule_ != nullptr && flags.curRule_ != r) continue;
      if (r == flags.curRule_) {
         if (cond_ || (flags.learnMode_ && learnCond_ != nullptr) 
             || (!flags.learnMode_ && detectCond_ != nullptr)) {
            bool needComma = false;
            os << "if (";
            if (cond_ != nullptr) {
               cond_->codeGen(os, indent+STEP_INDENT, flags);
               needComma = true;
            }
            if (flags.learnMode_ && learnCond_ != nullptr) {
               if (needComma) os << " && ";
               learnCond_->codeGen(os, indent+STEP_INDENT, flags);
            }
            if (!flags.learnMode_ && detectCond_ != nullptr) {
               if (needComma) os << " && ";
               detectCond_->codeGen(os, indent+STEP_INDENT, flags);
            }
            os << ") ";
         }
         if (wrt_ != nullptr && wrt_->size() > 0) 
            os << "__inc3(";
         else os << "__inc2(";
         os << name() << "_";
         if (separateStatsForEachRule_)
            os << to_string(i);
         else os << '0';
         if (wrt_ != nullptr && wrt_->size() > 0) {
            unsigned j=0;
            if (wrt_->size() > 1) 
               os << ", make_tuple(";
            else j = 1;
            for (auto e: *wrt_) {
               if (j++ != 0) os << ", ";
               e->codeGen(os, indent, flags);
            }
            if (wrt_->size() > 1) os << ")";
         }
         os << ", ";
         if (value_)
            value_->codeGen(os, indent, flags);
         else os << "0x1";
         os << ");"; 
         prtln(os, indent);
      }
      i++;
   }
}

void 
OnEntry::print(ostream &os, int indent) const {
  indent += STEP_INDENT;
  os << "on ";

  if (separateStatsForEachRule_) os << "eachof ";

  os << "{ ";
  for (unsigned i=0; i < trans_->size();) {
	os << (*trans_)[i]->label()->name();
	if (++i < trans_->size())
	  os << ", ";
  }
  os << "} ";
  
  if (cond_ != nullptr) {
    os << "when ";
    cond_->print(os, indent);
    os << ' ';
  }

  if (wrt_ != nullptr) {
      os << "specialize ("; unsigned j=0;
      for (auto it: *wrt_) {
         it->print(os, indent+STEP_INDENT);
         if (j++ != 0) os << ", ";
      }
      os << ") ";
	  if (size_ != nullptr) {
		os << "maxsize ";
		size_->print(os, indent);
		os << ' ';
	  }
  }
  
  if (minInterval_ != nullptr) {
    os << "histogram frequency ";
    minInterval_->print(os, indent);
    os << " [";
    numScales_->print(os, indent);
    os << "] ";
  }

  if (value_ != nullptr) {
    os << "histogram ";
	value_->print(os, indent);
    os << ' ';
  }

  if (maxprob_ != nullptr) {
	os << "maxprob ";
	maxprob_->print(os, indent);
    os << ' ';
  }

  if (learnCond_ != nullptr) {
	os << "learn if ";
	learnCond_->print(os, indent);
    os << ' ';
  }

  if (detectCond_ != nullptr) {
	os << "detect if ";
	detectCond_->print(os, indent);
  }
}

