// This is a copy fron nBMSL which seemed close enough, and possibly
// more clean-looking than the version that was originally here.
// Also matches the version from the 504/e-- directory (after reordering)

#ifndef STE_CLASSES_H
#define STE_CLASSES_H

#include "all.h"
#include "ST.h"
#include "Value.h"
#include "ProgramElem.h"
#include <set>

class StmtNode;
class RuleNode;
class ExprNode;
class PatNode;
class VariableEntry;
class OpNode;
class OnEntry;
class PrimitivePatNode;
class CompoundStmtNode;
class InvocationNode;
class RefExprNode;

class ExprNode ;
class SymTabEntry: public ProgramElem {
 public:
  static vector<SymTab*> symTabs_;

 private:
  string name_;
  SymKind kind_;
  SymTab* st_;
  // These following fields are used to link the STEs so that their order
  // of declaration can be maintained.
  SymTabEntry* next_;
  SymTabEntry* prev_;
  SymTab* parent_;

 public:
  /*****************************************************************
	The implementation of this class, as well its descendents.
	makes definite assumptions about the use of kind, name, prev, next
	and the symbol table fields. As such, methods operating on these
	fields are non-virtual. (Overriding them in subclasses is strongly
	discouraged.) All other operations are virtual, whether or not
	we can see any potential use of overriding them.
  ****************************************************************/

  SymTabEntry(string name, SymKind kind, int flags, 
              int line, int column, string file, const Type* t=nullptr);
  virtual ~SymTabEntry() {};

  string name() const { return name_; }
  SymKind kind() const { return kind_; } 
  const SymTab* symTab() const { return st_; };
  const SymTabEntry* prev()const { return prev_ ;};
  const SymTabEntry* next()const  { return next_ ;};

  void name(string str) { name_=str; };
  void kind(SymKind kind) { kind_ = kind; };
  SymTab* symTab() { return st_; };
  void prev(SymTabEntry *se) { prev_ = se; }
  SymTabEntry* prev() { return prev_ ;};
  void next(SymTabEntry *se) { next_ = se; }
  SymTabEntry* next()  { return next_ ;};
  void parent(SymTab *s) { parent_ = s; }
  SymTab* parent()  { return parent_ ;};

  ErrorST insert(SymTabEntry* ste) {
    if(st_ == nullptr) {
       st_ = new SymTab(this);
       symTabs_.push_back(st_);
    }
    return st_->insert(ste);
  }

  virtual ExprNode* storageAlloc();
  
  virtual bool operator==(const SymTabEntry&) const;
  virtual bool operator!=(const SymTabEntry& ste) const 
  { return !operator==(ste); };

  virtual void print(ostream& os, int indent=0)  const = 0;

  // emitST calls print on descendants if fl is null, otherwise calls codeGen
  virtual void emitST(ostream& os,int ind=0,char ldelim='{',char rdelim='}',
					   bool linebreaks=true, CGFlags* fl=nullptr,
                      int first=0, int last=0) const {
    if (symTab() != nullptr) 
 	  symTab()->emitST(os, ind, ldelim, rdelim, linebreaks, fl, first,last);
  }
  virtual void codeGen(ostream& os, int indent, CGFlags& flags) {
     /*
     SymTab *st = symTab();
     if (st != nullptr) 
        for (auto ste: *st)
           ste->codeGen(os, indent, flags);
     */
  }
  virtual const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes);
  virtual void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  
 private:
  const SymTabEntry& operator=(const SymTabEntry&);
  //SymTabEntry(const SymTabEntry&);
  //SymTabEntry(SymTabEntry*);
};
	
inline ostream& operator<<(ostream& out, const SymTabEntry* ste){
  ste->print(out, 0);
  return out;
}

/****************************************************************
  The first few classes dont really add any functionality over
  the base class, except for providing a convenient constructor.
****************************************************************/

class GlobalEntry: public SymTabEntry {
 public:
  GlobalEntry(string name, int fl=-1, int line=0, int column=0, string file=""):
    SymTabEntry(name, GLOBAL_KIND, fl, line, column,file) {};
  ~GlobalEntry() {};
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes);
  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream&, int indent=0) const;
};

class InterfaceEntry: public SymTabEntry {
 public:
  InterfaceEntry(string nm, int f=-1, int line=0, int column=0, string file=""):
    SymTabEntry(nm, INTERFACE_KIND, f, line,column, file) {};
  ~InterfaceEntry() {};

  void codeGen(ostream& os, int indent, CGFlags& flags) {};
  void print(ostream& os, int indent) const {};
};

class ClassEntry: public SymTabEntry {
 public:
  ClassEntry(string name, int fl=-1, int line=0, int column=0, string file="")
    : SymTabEntry(name, CLASS_KIND, fl, line,column, file) {};
  ~ClassEntry() {};

  void codeGen(ostream& os, int indent, CGFlags& flags) {};
  void print(ostream& os, int indent) const;
};

class BlockEntry: public SymTabEntry {
 public:
  BlockEntry(string name, int fl=-1, int ln=0, int col=0, string file=""):
    SymTabEntry(name, BLOCK_KIND, fl, ln, col, file, (Type*)&Type::voidType) {};
  ~BlockEntry() {};
  void codeGen(ostream& os, int indent, CGFlags& flags) 
    { errMsg("BlockEntry::codeGen not defined", this); };
  void print(ostream& out, int indent=0) const; 
};

class RuleBlockEntry: public BlockEntry {
 public:
  RuleBlockEntry(int fl=-1, int line=0, int column=0, string file=""):
    BlockEntry(newName("rule"), fl, line,column, file) { kind(RULE_BLOCK_KIND);};
  ~RuleBlockEntry() {};
};

class EventBlockEntry: public BlockEntry {
 public:
  EventBlockEntry(int fl=-1, int line=0, int column=0, string file=""):
    BlockEntry(newName("event"), fl, line,column, file) 
	{ kind(EVENT_BLOCK_KIND);};
  ~EventBlockEntry() {};
};

// For functions, const means they don't change argument values, or modify any
// global state. However, we don't have a way to check the "not modifying 
// global state" part. 

class FunctionEntry: public SymTabEntry {
 private:
  bool isConst_;
  int offset_;
  CompoundStmtNode* body_;
  bool usesModuleParams_;
  string code_; // If code already emitted, the name of fn in generated code. 

 public:
  FunctionEntry(string name, Type* type=nullptr, int offset=1, int fl=-1,
				int line=0, int column=0, string file="", bool isConst= false):
    SymTabEntry(name, FUNCTION_KIND, fl, line,column, file, type) {
    isConst_ = isConst;
    offset_ = offset; body_ = nullptr; usesModuleParams_ = false; // fix later
 };
  ~FunctionEntry() {};

  bool isConst() const { return isConst_;}
  void isConst(bool b) { isConst_ = b;};

  int offset() const { return offset_;}
  void offset(int o) { offset_ = o;}

  const CompoundStmtNode* body() const { return body_;};
  CompoundStmtNode* body() {return body_;};
  void body(CompoundStmtNode* n) { body_ = n;};

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent) const;
};

class EnumElemEntry: public SymTabEntry {
 private:
  int num_;

 public:
  EnumElemEntry(string nm, int n, Type*t=0, int fl=-1, 
                int ln=0, int col=0, string f=""):
    SymTabEntry(nm, ENUM_ELEM_KIND, fl, ln, col, f, t), num_ (n) {};
  ~EnumElemEntry() {};

  int num() const { return num_;};
  void num(int i) {num_= i;};

  void codeGen(ostream& os, int indent, CGFlags& flags) 
    { errMsg("EnumElemEntry::codeGen not defined", this); }
  void print(ostream& os, int indent) const { os << name();};
};

/****************************************************************
  Following classes add more significant functionality to that
  provided by SymTabEntry.
****************************************************************/

class LabelEntry: public SymTabEntry {
 private:
  RuleNode *rule_;

 public:
  LabelEntry(string nm, RuleNode *rn=nullptr, int fl=-1,
             int ln=0, int col=0, string f=""):
    SymTabEntry(nm, LABEL_KIND, fl, ln, col, f) { rule_=rn;};
  ~LabelEntry() {};

  const RuleNode* rule() const { return rule_;};
  RuleNode* rule() { return rule_;};

  void rule(RuleNode* rn) { rule_ = rn;};
  
  void codeGen(ostream& os, int indent, CGFlags& flags)
    { os << "/* " << name() << ": */"; }
  void print(ostream& os, int indent) const {internalErr("shouldnt call");};
  //const Type* typeCheck() const {internalErr("shouldnt call"); return nullptr;};
};

class EventEntry: public SymTabEntry {
 private:

 public:
  EventEntry(string name, int fl=-1, int line=0, int column=0, string file=""):
    SymTabEntry(name, EVENT_KIND, fl, line,column, file) {};
  ~EventEntry() {};

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& out, int indent=0) const; 
};  

class VariableEntry: public SymTabEntry {
 private:
  VarKind vkind_;
  ExprNode* offset_;
  ExprNode* initVal_;
  vector<ExprNode*> subscripts_;// used if variable is an array.

 public:
  VariableEntry(string name, VarKind v, Type* type=nullptr, ExprNode* init=0, 
                int fl=-1, int ln=0, int col=0, string file=""):
    SymTabEntry(name, VARIABLE_KIND, fl, ln, col, file, type) {
    vkind_ = v; offset_ = nullptr; initVal(init);
 };

  //VariableEntry(const VariableEntry &v);
  ~VariableEntry() {};

  VarKind varKind() const { return vkind_;};
  void varKind(VarKind v) { vkind_ = v;};

  const ExprNode* offset() const { return offset_;}
  ExprNode* offset() { return offset_;}
  void offset(ExprNode* offset) { offset_ = offset;};

  const ExprNode* initVal() const { return initVal_;}
  ExprNode* initVal() { return initVal_;};
  void initVal(ExprNode *init) { initVal_ = init;};

  unsigned int nSubscripts() const { return subscripts_.size();};
  const ExprNode* subscript(unsigned int i) const {
    if (subscripts_.size() > i) return subscripts_[i]; else return nullptr;
  }
  ExprNode* subscript(unsigned int i) {
    if (subscripts_.size() > i) return subscripts_[i]; else return nullptr;
  }
  void addSubscript(ExprNode* e) { subscripts_.push_back(e);}

  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes);

  // We want to identify variables that are (a) declared but not used, and 
  // (b) declared but not assigned. A helper function usedSyms, operating on all
  // STEClasses, is used to gather this information. 
  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent=0) const;
};

class EnumEntry: public SymTabEntry {
 private:
  vector<EnumElemEntry*>* elems_;

 public:
  EnumEntry(string name, int fl=-1, int ln=0, int column=0, string filename=""):
    SymTabEntry(name, ENUM_KIND, fl, ln,column, filename) {
    elems_ = new vector<EnumElemEntry*>();
 }
  //EnumEntry(const EnumEntry&);
  ~EnumEntry() {};

  const vector<EnumElemEntry*>* elems() const { return elems_;};
  int nelems() const { return elems_->size();}

  vector<EnumElemEntry*>* elems() {return elems_;};
  void addElem(EnumElemEntry* ste) {
    elems_->push_back(ste); 
 };
  
  //const Type* typeCheck();
  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent) const;

 private:
  EnumEntry& operator=(EnumEntry&);
};

class ModuleEntry: public SymTabEntry {
 private:
  bool isStateMachine_;
  EnumElemEntry *startState_, *finalState_;
  map<EnumElemEntry*, ExprNode*> timeout_;
  vector<PrimitivePatNode*> mapDecls_;
  /* Each map statement is like a primitive pattern. */
  vector<OnEntry*> onDecls_;
  vector<RuleNode*> rules_;
  vector<InterfaceEntry*> imports_;
  vector<InvocationNode*> instantiate_;

 public:
  ModuleEntry(string name, int fl=-1, int line=0, int column=0, string file="");
  ~ModuleEntry() {};

  bool isStateMachine() const { return isStateMachine_;};
  void isStateMachine(bool b) { isStateMachine_ = b;};

  const EnumElemEntry* startState() const { return startState_;};
  EnumElemEntry* startState() { return startState_;};
  void startState(EnumElemEntry* s) { 
    if (startState_ == nullptr) startState_ = s; 
    else errMsg("StartState already defined");
 };

  const EnumElemEntry* finalState() const { return finalState_;};
  EnumElemEntry* finalState() { return finalState_;};
  void finalState(EnumElemEntry* s) {
    if (finalState_ == nullptr) finalState_ = s; 
    else errMsg("FinalState already defined");
 };

  ExprNode* timeout(const EnumElemEntry* s) const {
    if ((s == nullptr) || (timeout_.count((EnumElemEntry*)(s)) == 0)) return 0;
    else return  timeout_.find((EnumElemEntry*)(s))->second; 
  };
  void timeout(EnumElemEntry* s, ExprNode* e) {
    if (timeout_.count(s) == 0) timeout_[s] = e;
    else if (timeout_[s] != e)
	  errMsg("Multiple timeouts declared for state '" + s->name() + "'");
  };
  void timeout(vector<EnumElemEntry*> *s, ExprNode* d) { 
	if (s != nullptr) 
       for (auto e: *s)
          timeout(e, d);
  }

  const vector<RuleNode*> rules() const { return rules_;};
  vector<RuleNode*> rules() { return rules_;};
  const RuleNode* rule(int i) const { return rules_[i];}
  RuleNode* rule(int i) { return rules_[i];}
  void addRule(RuleNode* re) { rules_.push_back(re);};

  void addImport(InterfaceEntry* ie) { imports_.push_back(ie);};

  void addMap(PrimitivePatNode* p) {if (p!=nullptr) mapDecls_.push_back(p);}; 
  void addOnEntry(OnEntry* e) {if (e != nullptr) onDecls_.push_back(e);}; 

  void addInstantiate(InvocationNode *in) { instantiate_.push_back(in); };
  
  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes);

  vector<RuleNode*>
     instantiate(vector<ExprNode*> params, CGFlags& flags) const;
  void emitDecls(ostream& os, int indent, CGFlags& flags);
  void emitVars(ostream& os, int indent, CGFlags& flags);
  void emitInit(ostream& os, int indent, CGFlags& flags);
  void codeGen(ostream& os, int indent, CGFlags& flags);
  void mainCodeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& os, int indent)const;

 private:
  //ModuleEntry(const ModuleEntry&);
  //ModuleEntry& operator=(const ModuleEntry&);
};

class StructEntry: public SymTabEntry {
 private: 
  vector<StructEntry*> parent_;
  vector<ExprNode*> constraint_;
  vector<StructEntry*> child_; 
  ExprNode* size_;

 public:
  StructEntry(string name, int fl=-1, int line=0, int column=0, string file="");
  ~StructEntry() {};

  unsigned int nParents() const { return parent_.size();};

  const StructEntry* parent(unsigned int i) const 
    { return (i < parent_.size()) ? parent_[i] : nullptr;};
  StructEntry* parent(int i)
    { return (StructEntry*)(((const StructEntry*)this)->parent(i));};
  void addParent(StructEntry* p) {
	if (p!=nullptr) {parent_.push_back(p); p->addChild(this);}}

  const ExprNode* constraint(unsigned int i) const 
    { return (i < constraint_.size()) ? constraint_[i] : nullptr;};
  ExprNode* constraint(int i)
    { return (ExprNode*)(((const StructEntry*)this)->constraint(i));};
  void addConstraint(ExprNode* c) {constraint_.push_back(c);};
  // Note: should NOT check if c==nullptr, since it is the same as c='true'

  const StructEntry* child(unsigned int i) const
    { return (i < child_.size())? child_[i] : nullptr;};
  StructEntry* child(int i)   { 
     return (StructEntry*)(((const StructEntry*)this)->child(i));};
  void addChild(StructEntry* se) {if (se != nullptr) child_.push_back(se);};

  const ExprNode* size() const { return size_;};
  ExprNode* size() { return size_;};

  //const Type* typeCheckInit(set<const SymTabEntry*>& avlParams, 
  //                      map<VarKind, bool>& valVtypes);
  //const Type* typeCheck();
  const SymTabEntry* lookUpWithInheritance(string var, 
										   const ProgramElem *pe=nullptr) const;
  const SymTabEntry* lookUpInDescendants(string v, list<StructEntry*>& parents,
										 const ProgramElem *pe=nullptr) const;
  const SymTabEntry* lookUpStruct(string var, RefExprNode* base, 
				  ExprNode*& constraints, list<StructEntry*> parents, 
				  const ProgramElem *pe) const;
  void specializeStruct(list<StructEntry*>, const ProgramElem*); // ????

  void codeGen(ostream& os, int indent, CGFlags& flags);
  void print(ostream& out, int indent=0) const; 
};

class OnEntry: public SymTabEntry {
 private:
  bool separateStatsForEachRule_;
  ExprNode *size_, *cond_, *maxprob_; // cond_ is obsolete?
  //ExprNode *decayrate_, *decaytime; // To be added?
  ExprNode *learnCond_, *detectCond_;

  ExprNode *minInterval_, *numScales_;
  ExprNode *value_;

  vector<RuleNode*>* trans_;
  vector<ExprNode*>* wrt_;

 public:
  OnEntry(string name, vector<RuleNode*>* trans, 
	  int fl=-1, int line=0, int column=0, string file="");
  ~OnEntry() {};

  const ExprNode* cond() const;
  const ExprNode* size() const;
  vector<RuleNode*> trans() const { return *trans_; };
  void trans(vector<RuleNode*>* t) { trans_ = t; };

  void onEachOf(bool f=true) { separateStatsForEachRule_ = f; };
  void size(ExprNode* s) { size_ = s;}
  void cond(ExprNode* cond) { cond_ = cond;}
  void maxprob(ExprNode* fa) { maxprob_ = fa;}
  void learnCond(ExprNode* e) { learnCond_=e; }
  void detectCond(ExprNode *e) { detectCond_=e; }

  void interval(ExprNode* tl, ExprNode* n) {minInterval_=tl; numScales_=n;};

  void value(ExprNode* v) { value_ = v;}

  ExprNode* learnCond() { return learnCond_; }
  ExprNode* detectCond() { return detectCond_; }

  void wrt(vector<ExprNode*>* wrts) { wrt_ = wrts;}

  void usedSyms(set<SymTabEntry*>& use, set<SymTabEntry*>& assign);
  const Type* typeCheck(set<const SymTabEntry*>& avlParams, 
                        map<VarKind, bool>& valVtypes);

  void codeGen(ostream& os, int indent, CGFlags& flags);
  vector<string> emitVars(ostream& os, int indent, CGFlags& flags);
  void emitInit(ostream &os, int indent, CGFlags& flags);
  void print(ostream &os, int indent) const;
};

#endif
