#include "Value.h"
#include "RegExpr.h"
#include "STEClasses.h"
#include <climits>

string vKindTag[] = {
   "FOREACH_VAR", "FN_PARAM_VAR", "LOCAL_VAR", "GLOBAL_VAR", "MODULE_PARAM_VAR",
   "EVENT_PARAM_VAR", "PATTERN_VAR" /* for REWBR*/, "RULE_VAR",
   "COMPILER_GENERATED_VAR",
   "UNDEFINED"
};

string varKindTag(VarKind vk) {
   assert(vk <= UNDEFINED);
   return vKindTag[vk];
}

const char* typeName[] = {
  "error", "unknown", "void", 
  "string", "regexpr",
  "bool", 
  "bit", 
  "unsigned char", "char", 
  "unsigned short", "short",
  "enum", 
  "unsigned int", "int", 
  "unsigned long", "long",
  "double",
  "struct", "class", 
  "vector", "set", "dict", "tuple", 
  "module", "event", "function",
  "numeric",
  "integral",
  "signed",
  "unsigned",
  "scalar",
  "primitive",
  "dict or set",
  "container",
  "container_or_str",
  "native",
  "data",
  "any"
};

const char* codename[] = {
  "ET_ERROR", "ET_UNKNOWN", "void", 
  "string", "regexpr",
  "bool", 
  "bool", "unsigned char", "char", // BIT type mapped to bool
  "unsigned short", "short",
  "enum", 
  "unsigned", "int", 
  "unsigned long", "long",
  "double",
  "ET_STRUCT", "ET_CLASS", 
  "vector", "unordered_set", "unordered_map", "tuple", 
  "ET_MODULE", "ET_EVENT", "ET_FUNCTION",
  "ET_NUMERIC",
  "ET_INTEGRAL",
  "ET_SIGNED",
  "ET_UNSIGNED",
  "ET_SCALAR",
  "ET_PRIMITIVE",
  "ET_DICTORSET",
  "ET_CONTAINER",
  "ET_CONTAINER_OR_STR",
  "ET_NATIVE",
  "ET_DATA",
  "ET_ANY"
};

/* 
   sizes of various types in number bits 
   negative sizes indicate that the types size is undefined.
*/
const int Type::SIZE_UNDEF = -1;
int typeSize[] = {
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF,
  Type::SIZE_UNDEF, Type::SIZE_UNDEF,
  1,
  1,          
  8,          8,
  16,         16,
  32,
  32,         32,
  64,         64,
  64,
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF,
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF, 
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF,
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF,
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF,
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF,
  Type::SIZE_UNDEF, Type::SIZE_UNDEF, Type::SIZE_UNDEF
};

unsigned long maxVals[] = {
            0, 0, 0,
            0, 0,
            1,         
            1,   
            UCHAR_MAX, CHAR_MAX,
            USHRT_MAX, SHRT_MAX,
            INT_MAX,
            UINT_MAX,  INT_MAX,
            ULONG_MAX, LONG_MAX,
            0,
            0, 0, 0,
            0, 0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            0, 0
};

long minVals[] = {
            0, 0, 0,
            0, 0,
            0,         
            0,
            0, CHAR_MIN,
            0, SHRT_MIN,
            0, 
            0, INT_MIN,
            0, LONG_MIN,
            0,
            0, 0, 0,
            0, 0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            0, 0, 0,
            0, 0, 0 ,
            0, 0
};

const string Type::name(TypeTag t) {
  if ((t >= ERROR) && (t <= ANY))
    return (t.isConst()? "const " : "") + string(typeName[t.tag()]);
  else return string();
}

const string Type::codeName(TypeTag t, bool inclConst) {
  if((t >= ERROR) && (t <= ANY))
    return (inclConst && t.isConst()? "const " : "") + string(codename[t.tag()]);
  else return string();
}

int Type::size(BaseTypeTag t) {
  if((t >= ERROR) && (t <= ANY))
    return (typeSize[t]);
  else return Type::SIZE_UNDEF;
}

const Type
  Type::errorType(ERROR), Type::unkType(UNKNOWN), Type::voidType(VOID),
  Type::stringType(STRING), Type::constStringType(TypeTag(STRING, true)), 
  Type::regExprType(REGEXPR),
  Type::boolType(BOOL),
  Type::bitType(BIT), 
  Type::ucharType(UCHAR), Type::charType(CHAR), 
  Type::ushortType(USHORT), Type::shortType(SHORT), 
  Type::uintType(UINT), Type::intType(INT), 
  Type::ulongType(ULONG), Type::longType(LONG), 
  Type::doubleType(DOUBLE);

const Type
  Type::constUintType(Type::TypeTag(UINT, true));

const Type* Type::type[] = {
  &Type::errorType, &Type::unkType, &Type::voidType, 
  &Type::stringType, &Type::regExprType,
  &Type::boolType,
  &Type::bitType, 
  &Type::ucharType, &Type::charType, 
  &Type::ushortType, &Type::shortType, 
  nullptr, 
  &Type::uintType, &Type::intType,
  &Type::ulongType,  &Type::longType,
  &Type::doubleType,
  nullptr, nullptr, nullptr,
  nullptr, nullptr, nullptr, nullptr,
  nullptr, nullptr, nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr,
  nullptr
};

/****************************************************************
   Type Class 
****************************************************************/

Type::Type(TypeTag tag) {
  tag_ = tag; 
  argTypes_ = nullptr; 
  retType_ = nullptr;
};     

Type::Type(SymTabEntry* td, TypeTag t) {   // Must be enum, class or struct
  tag_ = t; 
  typeDesc_ = td; 
};

Type::Type(Type* elemType, TypeTag t) { // For list and set types
   assert(t == VECTOR || t == SET);
   tag_ = t;
   elemType_ = elemType; 
};

Type::Type(Type* keyType, Type* elemType) { // For dict type
   tag_ = DICT;
   keyType_ = keyType;
   elemType_ = elemType; 
};

Type::Type(vector<Type*>* tupleType, TypeTag t) {// For tuples, modules, events
  tag_ = t; 
  argTypes(tupleType); 
};

Type::Type(vector<Type*>* argt, Type* rt) {      // For functions
  tag_ = FUNCTION; 
  retType_ = rt; 
  argTypes_ = argt; 
};

string 
Type::fullName() const {
  if (!this) 
     return "nullptr";
  string info;
  if (isConst())
	info += "const ";
  switch (tag()) {
  case ENUM:
  case STRUCT:
  case CLASS:
     return name() + " " + typeDesc_->name() ;
     
  case VECTOR:
     return info + elemType_->fullName() + " "+ name() ;

  case SET:
     return info + elemType_->fullName() + " "+ name() ;

  case DICT:
     return info + name() + "(" + keyType_->fullName() + ", " +
              elemType_->fullName() + ") ";

  case TUPLE:
  case MODULE:
  case EVENT:
  case FUNCTION:
     info += name(tag_) + " (";
     if (argTypes_ != nullptr) {
        for (unsigned int i=0; i< argTypes_->size(); i++) {
           if ((*argTypes_)[i] != nullptr)
              info += (*argTypes_)[i]->fullName() ;  
           if (i < argTypes_->size() - 1)
              info += ", ";
        }
     }
     if (tag_ == FUNCTION) {
		info += "): ";
		info += retType_->fullName() + " ";
     }
     else info += ") ";
     return info;

  default:
     if (info == "const " && name().find("const") < 2)
        info = "";
     break;
  }
  return info+name();
}

string 
Type::fullCodeName(bool inclConst) const {
  string info;
  switch (tag()) {
  case ENUM:
  case STRUCT:
  case CLASS:
     return codeName(inclConst) + " " + typeDesc_->name() ;
     
  case VECTOR:
     return codeName(inclConst)+ "<" +  elemType_->fullCodeName(inclConst)+ "> ";

  case SET:
     return codeName(inclConst)+ "<" +  elemType_->fullCodeName(inclConst)+ "> ";

  case DICT:
     return codeName(inclConst) + "<" +  keyType_->fullCodeName(inclConst) 
        + ", " + elemType_->fullCodeName() + "> ";

  case TUPLE: {
     info = codeName(inclConst) + "<";
     for (unsigned int i=0; i< argTypes_->size(); i++) {
        if ((*argTypes_)[i] != nullptr) 
           info += (*argTypes_)[i]->fullCodeName(inclConst);  
        if (i < argTypes_->size() - 1)
           info += ", ";
     }
     info += "> ";
     return info;
  }

  default:
     //internalErr("fullCodeName called on invalid type");
     return codeName();
  }
}

const Type&
Type::operator=(const Type& t) {
  tag_ = t.tag_;
  switch (tag()) {
    case ENUM:
    case STRUCT:
    case CLASS: 
      typeDesc_ = t.typeDesc_;
      break;

    case DICT:
      if (t.keyType_ != nullptr)
        keyType_ = new Type(*(t.keyType_));
	  else keyType_ = new Type(VOID);
    case SET:
    case VECTOR:
	  if (t.elemType_ != nullptr) 
		elemType_ = new Type(*(t.elemType_));
	  else elemType_ = new Type(VOID);
      break;

    case TUPLE:
    case MODULE:
    case EVENT:
    case FUNCTION: 
      if (t.argTypes_ != nullptr) {
	argTypes_ = new vector<Type*>(t.argTypes_->size());
	for (unsigned int i=0; i<t.argTypes_->size(); i++) 
	  (*argTypes_)[i] = new Type(*((*t.argTypes_)[i]));
      } 
	  else argTypes_ = nullptr;
	  if (tag_ == FUNCTION) {
		if (t.retType_ != nullptr)
		  retType_ = new Type(*(t.retType_));
		else retType_ = new Type(VOID);
      }
      break;

    default:
      break;
    }
  return *this;
}

void 
Type::print(ostream& os, int indent) const {
  os << fullName();
}

static bool
isSubType(vector<Type*>* l1, vector<Type*> *l2) { //Is *l1 a subType of *l2
  unsigned int i=0;
  if ((l1 == nullptr) || (l2 == nullptr))
    return (l1 == l2);
  else if (l1->size() != l2->size())
    return false;
  else {
    for (i=0; (i < l1->size()); i++)
      if (!(*l1)[i]->isSubType(*(*l2)[i])) 
         return false;
	return true;
  }
}

bool 
Type::isSubType(const Type& t, bool relaxed) const {
  if (tag() <= BOOL)
    return (tag() == t.tag());
  else if (isNumeric(tag())) {
     if (!isNumeric(t.tag()))
        return false;
     if (tag() <= t.tag() && (isSigned(t.tag()) || !isSigned(tag())))
        return true;
     return (relaxed && (!isIntegral(t.tag()) || isIntegral(tag())));
  }
  else if (tag() != t.tag())
    return false;
  else {
    switch (tag()) {
    case ENUM:
    case STRUCT:// name equivalence, for subtyping follow inheritance chains.
	  // However, it is not clear that we need to support inheritance based
	  // subtyping in the language, as its utility is unclear. So, we
	  // simplify things by using name equivalence only.
    case CLASS: // name equivalence, subtype if equivalent
      return (typeDesc_ == t.typeDesc_);

    case SET:
    case VECTOR:  // structural equivalence and subtyping
      return elemType_->isSubType(*t.elemType_) || elemType_->tag() == NATIVE;
      // Note: NATIVE tag will be inferred only for empty lists. Since 
      // empty lists are assignable to arbitrary lists, we shd return true.

    case DICT:
      return (elemType_->isSubType(*t.elemType_) && 
              keyType_->isSubType(*t.keyType_));

    case FUNCTION: // structural equivalence and subtyping
      return (retType_->isSubType(*t.retType_) && 
			  ::isSubType(t.argTypes_, argTypes_));

    case TUPLE:
    case MODULE:
    case EVENT:
      return ::isSubType(argTypes_, t.argTypes_);

    default:
      break;
    }
  }
  internalErr("Type::isSubType: unhandled case");
  return false;
}

bool
Type::isSubType(TypeTag t) const {
  if (Type::type[t.tag()] != nullptr) {
    return isSubType(*type[t.tag()]);
  }
  else if (tag() == t.tag())
     return true;
  else {
    switch (t.tag()) {
    case VECTOR:
      return isVector(tag());
    case SET:
      return isSet(tag());
    case DICT:
      return isDict(tag());
    case TUPLE:
      return isTuple(tag());

    case NUMERIC:
      return isNumeric(tag());
    case INTEGRAL:
      return isIntegral(tag());
    case SIGNED:
      return isSigned(tag());
    case UNSIGNED:
      return isUnsigned(tag());
    case SCALAR:
      return isScalar(tag());
    case PRIMITIVE:
      return isPrimitive(tag());
    case DICTORSET:
       return isDictOrSet(tag());
    case CONTAINER:
       return isContainer(tag());
    case CONTAINER_OR_STR:
       return isContainerOrStr(tag());
    case NATIVE:
      return isNative(tag());
    case DATA:
      return isData(tag());
    case ANY:
      return isValid(tag());
    default:
      cerr << name(tag()) << "  " << name(t) << endl ;
      internalErr("Unexpected case in Type::isSubType(TypeTag)");
      return false;
    }
  }
}

void Type::
unify(const Type* t, const ProgramElem* pe) {
   // NATIVE type is inferred only in case of empty containers. In such cases,
   // the result should be the more specific type.
   if (tag() == NATIVE) {
      *this = *t;
      return;
   }
   else if (t->tag() == NATIVE)
      return;

   if (isEquiv(*t))
      return;
   if (Type::isNumeric(tag()) && Type::isNumeric(t->tag())) {
      if (t->isSubType(*this))
         return;
      else if (isSubType(*t))
         *this = *t;
      else { // %%%% RELYING on successive tags for numeric types
         for (BaseTypeTag tn = tag(); tn <= DOUBLE; tn = (BaseTypeTag)(tn+1)) {
            *this = Type(tn);
            if (t->isSubType(*this))
               return;
         }
         warnMsg("Possible type error: types " + name() +
                 " and " + t->name() + " are incompatible", pe);
         *this = errorType;
      }
      return;
   }

   if (tag() != t->tag()) {
      warnMsg(string("Possible type error: types ") + typeName[tag()] 
           + " and " + typeName[t->tag()] + " are incompatible", pe);
      *this = errorType;
      return;
   }

   switch (tag()) {
     case TUPLE:
        if (t->argTypes()->size() != argTypes()->size()) {
           warnMsg(string("Possible error: tuples of different sizes are ") +
                   "incompatible", pe);
           *this = errorType;
        }
        else {
           vector<Type*>* at = argTypes();
           for (unsigned i=0; i < at->size(); i++) {
              (*at)[i]->unify((*t->argTypes())[i], pe);
              if (*(*at)[i] == errorType) {
                 *this = errorType;
                 return;
              }
           }
        }
        break;

     case DICT:
        keyType()->unify(t->keyType(), pe);
     case VECTOR:
     case SET:
        elemType()->unify(t->elemType(), pe);
        break;

     default: 
        *this = errorType;
        break;
   }
}


// == is a strict equality called from the equality of symbol table entries.
bool 
Type::operator==(const Type& t) const {
  if (this == &t) 
    return true;
  else if (tag() != t.tag())
    return false;
  else if (isTagOnly(tag()))
    return true;
  else {
    switch (tag()) {
    case ENUM:
    case STRUCT:
    case CLASS: 
      return (typeDesc_ == t.typeDesc_ ||
              (typeDesc_ && t.typeDesc_ && *typeDesc_ == *t.typeDesc_));

    case VECTOR:
    case SET:
      return (elemType_ == t.elemType_ ||
              (elemType_ && t.elemType_ && *elemType_ == *t.elemType_));

    case DICT:
      return ((elemType_ == t.elemType_ ||
               (elemType_ && t.elemType_ && *elemType_ == *t.elemType_)) &&
              (keyType_ == t.keyType_ ||
               (keyType_ && t.keyType_ && *keyType_ == *t.keyType_)));

    case FUNCTION: 
      if ((retType_ != t.retType_) &&
		  ((retType_ == nullptr) || (t.retType_==nullptr) ||
		   (*retType_ != *t.retType_)))
	return false;
      // fall thru to following cases is intentional

    case TUPLE:
    case MODULE:
    case EVENT:
      if (argTypes_ == t.argTypes_)
	return true;
      else return ((argTypes_ != nullptr) && (t.argTypes_ != nullptr) &&
		   (*argTypes_ == *t.argTypes_));
    default:
      break;
    }
  }
  internalErr("Type::operator==: unhandled case");
  return false;
}


/****************************************************************
     Value Class 
****************************************************************/


vector<Value*>*
copyValList(const vector<const Value*>& l) {
  vector<Value*>* vl = new vector<Value*>((vector<Value*>&)l);
  for (unsigned i= 0; i<l.size(); i++) { // Make deep copies of vector elements
	Value * v = (*vl)[i];
	(*vl)[i] = new Value(*v);
  }
  return vl;
}

Value::Value(Type::BaseTypeTag t, unsigned long ul) {
   if (!Type::isIntegral(t) && !Type::isBool(t))
      type_ = &Type::errorType;
   else {
      type_ = Type::type[t];
      lVal_ = (long)ul;
   }
}

Value::Value(long l) {
   lVal_ = l; type_ = &Type::errorType;
   for (unsigned i = Type::BIT; i <= Type::LONG; i++)
      if (i == Type::ENUM) continue;
      else if (!Type::isSigned((Type::BaseTypeTag)i) && l < 0) continue;
      else if (l >= minVals[i] && l <= (long)maxVals[i]) {
         type_ = Type::type[i];
         break;
      }
}

Value::Value(unsigned long ul) {
   lVal_ = (long)ul; type_ = &Type::errorType;
   for (unsigned i = Type::BIT; i <= Type::LONG; i++)
      if (i == Type::ENUM) continue;
      else if (ul <= maxVals[i]) {
         type_ = Type::type[i];
         break;
      }
}

Value::Value(unsigned ul) {
   lVal_ = (long)ul; type_ = &Type::errorType;
   for (unsigned i = Type::BIT; i <= Type::LONG; i++)
      if (i == Type::ENUM) continue;
      else if (ul <= maxVals[i]) {
         type_ = Type::type[i];
         break;
      }
}

Value::Value(const RegExpr* re) { 
reVal_ = new RegExpr(*re); 
type_ = &Type::regExprType; 
}

Value::Value(int i, SymTabEntry* t) {
  if (t != nullptr && t->kind() == ENUM_ELEM_KIND) {
    type_ = t->type();
    lVal_ = i ;  
  } 
  else {
    internalErr("Value::Value: Expected enumerated type as argument");
    type_ = &Type::errorType ;
  }
}

Value::Value(const vector<Value*>* l, Type *t)   { // for lists+tuples
   if ((t) && (t->tag() == Type::VECTOR || t->tag() == Type::TUPLE)) {
      type_ = t; 
      listVal_ = copyValList(*(const vector<const Value*>*)l);
   }
   else {
      type_ = &Type::errorType ;
      internalErr("Value should be a list or tuple type");
   }
}

// **** Invariant for all constructors: type_ is non-null and correctly set,
// **** and appropriate union field is properly initialized.

void
Value::releaseResources() {
   if (isString() && sVal_) {
      delete sVal_;
      sVal_ = nullptr;
   }
   else if (isRegExpr() && reVal_) {
      delete reVal_;
      reVal_ = nullptr;
   }
   else if (type_->tag() == Type::VECTOR || type_->tag() == Type::TUPLE) {
      if (listVal_) {
         for (auto v: *listVal_)
            if (v) delete v;
         delete listVal_;
         listVal_ = nullptr;
      }
   }
}

Value::~Value() {
  releaseResources();
}

void static prtStrVal(string& s, ostream& os) {
   for (unsigned char c: s)
      if (c == '"')
         os << '\\' << '"';
      else if (c == '\\')
         os << '\\' << '\\';
      else if (c >= ' ' && c < 128)
         os << c;
      else {
         char b0 = (c & 0x7) + '0'; c >>= 3;
         char b1 = (c & 0x7) + '0'; c >>= 3;
         char b2 = (c & 0x7) + '0'; c >>= 3;
         os << '\\' << b2 << b1 << b0;
      }
}

void 
Value::codeGen(ostream& os, const Type* coercedType, string prefix, 
               int indent) const{

  Type::BaseTypeTag t = type_->tag();
  if (coercedType) {
     Type::BaseTypeTag ct = coercedType->tag();
     switch (ct) {
        case Type::UINT: os << ulval() << 'U'; return;
        case Type::INT: os << lval(); return;
        case Type::ULONG: os << ulval() << "UL"; return;
        case Type::LONG: os << lval() << 'L'; return;
        default:
           if (ct >= Type::BIT && ct <= Type::LONG) {
              os << '(' << coercedType->fullCodeName() << ')';
              if (Type::isSigned(ct))
                 os << lval(); 
              else os << ulval();
              return;
           }
           break;
     }
  }
  switch (t) {
    case Type::ERROR: 
       os << "EV_ERROR_VALUE "; break;
    case Type::VOID: 
       os << "EV_VOID_VALUE"; break;
    case Type::BOOL: 
       os << ((lVal_ == false) ? "false" : "true"); 
       break;
    case Type::ENUM: 
       os << prefix;
       (*((EnumEntry*)(type_->typeDesc()))->elems())[lVal_]->print(os, indent);
       break;

    default: 
       if (Type::isString(type_->tag())) {
          os << "string(\"";
          if (sVal_ != nullptr)
             prtStrVal(*sVal_, os);
          os << "\")";
       }
       else if (Type::isRegExpr(type_->tag())) {
          os << "regexpr(\"";
          if (reVal_ != nullptr)
             os << *reVal_;
          os << "\")";
       }
       else {
          os << '(' << (coercedType? coercedType : type())->fullCodeName() << ')';
          if (Type::isIntegral(t)) {
             if (Type::isSigned(t))
                os << lval(); 
             else os << ulval();
          }
          else if (Type::isDouble(t))
             os << dval();
          else internalErr("Value::print: unsupported value type");
       }
  }
}

void 
Value::print(ostream& os, int indent) const{
  if (!this) {
     os << "nullptr"; return;
  }
  Type::BaseTypeTag t;
  switch (t = type_->tag()) {
    case Type::ERROR: 
      os << "ErrorValue "; break;
    case Type::VOID: 
      os << "VoidValue"; break;
    case Type::BOOL: 
      os << ((lVal_ == false) ? "false" : "true"); 
      break;
    case Type::ENUM: 
       (*((EnumEntry*)(type_->typeDesc()))->elems())[lVal_]->print(os, indent);
       //os << ((EnumEntry*)(type_->typeDesc()))->name();
      break;
    case Type::VECTOR:
    case Type::TUPLE:
      char c1, c2;
      unsigned int i, n ;
      n = (listVal_ != nullptr) ? listVal_->size() : 0;
	if (t == Type::VECTOR) {
	c1 = '{'; c2 = '}';
      }
      else {
	c1 = '('; c2 = ')'; 
      }
      os << c1;
      for (i=0 ; i < n ;i++) {
	(*listVal_)[i]->print(os, indent); 
	if (i < n-1)
	  os << ", ";
      }
      os << c2;
      break;
    default: 
	if (Type::isString(type_->tag()))
	  if (sVal_ == nullptr)
		os << "nullptr POINTER";
	  else os << "\"" << *sVal_ << "\""; 
	else if (Type::isRegExpr(type_->tag()))
	  if (reVal_ == nullptr)
		os << "nullptr POINTER";
	  else os << "\"" << *reVal_ << "\""; 
	else if (Type::isIntegral(t))
	  if (Type::isSigned(t))
		os << lVal_; 
	  else os << (unsigned long)lVal_;
	else if (Type::isDouble(t)) 
	  os << dVal_;
	else internalErr("Value::print: unsupported value type");
  }
}

extern string newName(string type);

string 
Value::sval() const {
  if ((Type::isString(type_->tag())) && (sVal_ != nullptr))
      return *sVal_ ;
  else {
    internalErr("Value::sval: Incorrect Type ");
    return string();
  }
}

const RegExpr*
Value::reval() const {
  if (Type::isRegExpr(type_->tag()))
      return reVal_ ;
  else return nullptr;
}

RegExpr*
Value::reval() {
  if (Type::isRegExpr(type_->tag()))
      return reVal_ ;
  else return nullptr;
}

bool 
Value::bval() const {
  if (type_->tag() != Type::BOOL)
	internalErr("Value::bval: stored type is not bool");
  return lVal_;
}

/*
int 
Value::ival() const {  // for bit, byte, unsigned/signed short/int, long
  if (Type::isInt(type_->tag())) 
    return lVal_ ;
  internalErr("Value::ival: Incorrect Type ");
  return 0;
}
*/

long 
Value::lval() const { // for all integral types
  Type::BaseTypeTag tg = type_->tag();
  if (Type::isIntegral(tg))
     if (Type::isSigned(tg) || lVal_ <= ((long)maxVals[Type::LONG]))
        return lVal_;
  internalErr("Value::lval: Incorrect Type ");
  return 0;
}

unsigned long 
Value::ulval() const { // for all integral types
  Type::BaseTypeTag tg = type_->tag();
  if (Type::isIntegral(tg) && !Type::isSigned(tg))
     return (unsigned long)lVal_;

  internalErr("Value::lval: Incorrect Type ");
  return 0;
}

double
Value::dval() const { 
  if (Type::isDouble(type_->tag()))
    return dVal_; 
  else if (Type::isIntegral(type_->tag()))
     return (double)lval();
  else {
    internalErr("Value::dval: Incorrect Type ");
	return 0.0;
  }
}

const vector<const Value*>* 
Value::listVal() const { // for tuples, lists
  if (type_->tag() == Type::VECTOR || type_->tag() == Type::TUPLE)
    return (const vector<const Value*>*)listVal_;
  else {
    internalErr("Value::listVal: Incorrect Type ");
  return nullptr;
}
}

void
Value::sval(const string& s) { 
  if (Type::isString(type_->tag())) {
	if  (sVal_ != nullptr)
	   delete sVal_;
	sVal_ = new string(s);
  }
  else internalErr("Value::sval: Incorrect Type ");
}

void
Value::reval(const RegExpr& re) { 
  if (Type::isRegExpr(type_->tag())) {
	if  (reVal_ != nullptr)
	   delete reVal_;
	reVal_ = new RegExpr(re);
  }
  else internalErr("Value::reval: Incorrect Type ");
}

void 
Value::dval(double d) {
  if (Type::isDouble(type_->tag()))
    dVal_ = d;
  else internalErr("Value::dval: Incorrect Type ");
}

void 
Value::listVal(const vector<Value*>* vl) {
  if (type_->tag() == Type::VECTOR || type_->tag() == Type::TUPLE)
    listVal_ = copyValList(*(const vector<const Value*>*)vl);
  else internalErr("Value::listVal: Incorrect Type ");
}

/*******************************/
 
bool
Value::operator==(const Value& v) const {
  if ((*type_ == *v.type_) || 
	  (Type::isIntegral(type_->tag()) && Type::isIntegral(v.type_->tag()))) {
	Type::BaseTypeTag t = type_->tag();
	switch (t) {
	case Type::ERROR:
	case Type::VOID:
	  return true;
	case Type::ENUM:
	  return (lVal_ == v.lVal_);
	case Type::VECTOR:
	case Type::TUPLE:
	  return ((listVal_ == v.listVal_) ||
			  ((listVal_ != nullptr) && (v.listVal_ != nullptr) && 
			   (*listVal_ == *v.listVal_)));

	default:
	  if (Type::isString(t))
		return((sVal_ == v.sVal_) || 
               ((sVal_!=nullptr) && (v.sVal_!=nullptr) && (*sVal_ == *v.sVal_)));
	  else if (Type::isRegExpr(t))
	    return((reVal_ == v.reVal_) || 
               ((reVal_!=nullptr)&&(v.reVal_!=nullptr)&&(*reVal_ == *v.reVal_)));
      else if (Type::isBool(t) || Type::isIntegral(t))
		return (lVal_ == v.lVal_);
	  else if (Type::isDouble(t))
		return (dVal_ == v.dVal_);
	}
  }
  //internalErr("Value::operator==: Invalid argument types");
  return false;
}

bool
Value::operator<(const Value& v) const {
   Type::BaseTypeTag t = type_->tag(), vt = v.type_->tag();
   if (t != vt && (!isNumeric() || !v.isNumeric()))
      return (t < v.type_->tag());
   if (Type::isString(t))
      return (*sVal_ < *v.sVal_);
   else if (Type::isRegExpr(t))
      return (*reVal_ < *v.reVal_);
   else if (Type::isBool(t) || Type::isIntegral(t)) {
      if (Type::isSigned(t))
		return ((unsigned long)lVal_ < (unsigned long)v.lVal_);
      else return (lVal_ < v.lVal_);
   }
   else if (Type::isDouble(t))
      return (dVal_ < v.dVal_);
   else internalErr("Value::operator<: Invalid argument types");
  return false;
}

/********************************/

const Value&
Value::operator=(const Value& v) {
  releaseResources(); 
  type_ = v.type_; sVal_ = nullptr;
  //if ((*type_ == *v.type_) || 
  //	  (Type::isIntegral(type_->tag()) && Type::isIntegral(v.type_->tag()))) {
	Type::BaseTypeTag t = type_->tag();
	switch (t) {
	case Type::ERROR:
	case Type::VOID:
	  break;
	case Type::ENUM:
	  lVal_ = v.lVal_; break;
	case Type::VECTOR:
	case Type::TUPLE:
	  listVal_=copyValList(*(const vector<const Value*>*)(v.listVal_)); break;

	default:
	  if (Type::isString(t))
		sval(*v.sVal_);
	  else if (Type::isRegExpr(t))
		reval(*v.reVal_);
	  else if (Type::isBool(t) || Type::isIntegral(t))
		lVal_ = v.lVal_;
	  else if (Type::isDouble(t))
		dVal_ = v.dVal_;
	  else internalErr("Value::operator=: Invalid argument types");
	}
  // }
  return (*this);
}

size_t Value::
hash() const {
   if (isIntegral()) {
      long l = lval();
      return fasthash64(&l, sizeof(l), sizeof(l));
   }
   if (isDouble()) {
      double d = dval();
      return fasthash64(&d, sizeof(d), sizeof(d));
   }
   if (isString()) {
      const string& s = sval();
      const char* sc = s.data();
      return fasthash64(sc, s.size(), s.size());
   }
   if (isRegExpr()) {
      const RegExpr& re = *reval();
      return re.hash();
   }
   assert(false);
}
